import numpy
from nabu.io.cast_volume import (
    cast_volume,
    clamp_and_rescale_data,
    find_histogram,
    get_default_output_volume,
)
from tomoscan.esrf.volume import (
    EDFVolume,
    HDF5Volume,
    JP2KVolume,
    MultiTIFFVolume,
    TIFFVolume,
)
from nabu.io.writer import __have_jp2k__
from tomoscan.esrf.scan.edfscan import EDFTomoScan
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
import pytest
import h5py
import os
from silx.io.url import DataUrl


@pytest.mark.skipif(not __have_jp2k__, reason="need jp2k (glymur) for this test")
def test_get_default_output_volume():
    """
    insure nabu.io.cast_volume is working properly
    """
    with pytest.raises(TypeError):
        # test input_volume type
        get_default_output_volume(input_volume="dsad/dsad/", output_type="jp2")

    with pytest.raises(ValueError):
        # test output value
        get_default_output_volume(input_volume=EDFVolume(folder="test"), output_type="toto")

    # test edf to jp2
    input_volume = EDFVolume(
        folder="/path/to/my_folder",
    )
    output_volume = get_default_output_volume(
        input_volume=input_volume,
        output_type="jp2",
    )
    assert isinstance(output_volume, JP2KVolume)
    assert output_volume.data_url.file_path() == "/path/to/vol_cast"
    assert output_volume.get_volume_basename() == "my_folder"

    # test hdf5 to tiff
    input_volume = HDF5Volume(
        file_path="/path/to/my_file.hdf5",
        data_path="entry0012",
    )

    output_volume = get_default_output_volume(
        input_volume=input_volume,
        output_type="tiff",
    )
    assert isinstance(output_volume, TIFFVolume)
    assert output_volume.data_url.file_path() == "/path/to/vol_cast/my_file"
    assert output_volume.get_volume_basename() == "my_file"

    # test Multitiff to hdf5
    input_volume = MultiTIFFVolume(
        file_path="my_file.tiff",
    )
    output_volume = get_default_output_volume(
        input_volume=input_volume,
        output_type="hdf5",
    )
    assert isinstance(output_volume, HDF5Volume)
    assert output_volume.data_url.file_path() == "vol_cast/my_file.hdf5"
    assert output_volume.data_url.data_path() == "volume/" + HDF5Volume.DATA_DATASET_NAME
    assert output_volume.metadata_url.file_path() == "vol_cast/my_file.hdf5"
    assert output_volume.metadata_url.data_path() == "volume/" + HDF5Volume.METADATA_GROUP_NAME

    # test jp2 to hdf5
    input_volume = JP2KVolume(
        folder="folder",
        volume_basename="basename",
    )
    output_volume = get_default_output_volume(
        input_volume=input_volume,
        output_type="hdf5",
    )
    assert isinstance(output_volume, HDF5Volume)
    assert output_volume.data_url.file_path() == "folder/vol_cast/basename.hdf5"
    assert output_volume.data_url.data_path() == f"/volume/{HDF5Volume.DATA_DATASET_NAME}"
    assert output_volume.metadata_url.file_path() == "folder/vol_cast/basename.hdf5"
    assert output_volume.metadata_url.data_path() == f"/volume/{HDF5Volume.METADATA_GROUP_NAME}"


def test_find_histogram_hdf5_volume(tmp_path):
    """
    test find_histogram function with hdf5 volume
    """
    h5_file = os.path.join(tmp_path, "test_file")
    with h5py.File(h5_file, mode="w") as h5f:
        h5f.require_group("myentry/histogram/results/data")

    # if volume url provided then can find it
    assert find_histogram(volume=HDF5Volume(file_path=h5_file, data_path="myentry")) == DataUrl(
        file_path=h5_file,
        data_path="myentry/histogram/results/data",
        scheme="silx",
    )

    assert find_histogram(volume=HDF5Volume(file_path=h5_file, data_path="entry")) == None


def test_find_histogram_single_frame_volume(tmp_path):
    """
    test find_histogram function with single frame volume

    TODO: improve: for now histogram file are created manually. If this can be more coupled with the "real" histogram generation it would be way better
    """
    # create volume and histogram
    volume = EDFVolume(
        folder=tmp_path,
        volume_basename="volume",
    )
    histogram_file = os.path.join(tmp_path, "volume_histogram.hdf5")
    with h5py.File(histogram_file, mode="w") as h5f:
        h5f.require_group("entry/histogram/results/data")

    # check behavior
    assert find_histogram(volume=volume) == DataUrl(
        file_path=histogram_file,
        data_path="entry/histogram/results/data",
        scheme="silx",
    )

    assert find_histogram(
        volume=volume,
        scan=EDFTomoScan(scan=str(tmp_path)),
    ) == DataUrl(
        file_path=histogram_file,
        data_path="entry/histogram/results/data",
        scheme="silx",
    )

    assert find_histogram(
        volume=volume,
        scan=NXtomoScan(scan=str(tmp_path), entry="entry"),
    ) == DataUrl(
        file_path=histogram_file,
        data_path="entry/histogram/results/data",
        scheme="silx",
    )


def test_find_histogram_multi_tiff_volume(tmp_path):
    """
    test find_histogram function with multi tiff frame volume

    TODO: improve: for now histogram file are created manually. If this can be more coupled with the "real" histogram generation it would be way better
    """
    # create volume and histogram
    tiff_file = os.path.join(tmp_path, "my_tiff.tif")
    volume = MultiTIFFVolume(
        file_path=tiff_file,
    )
    histogram_file = os.path.join(tmp_path, "my_tiff_histogram.hdf5")
    with h5py.File(histogram_file, mode="w") as h5f:
        h5f.require_group("entry/histogram/results/data")

    # check behavior
    assert find_histogram(volume=volume) == DataUrl(
        file_path=histogram_file,
        data_path="entry/histogram/results/data",
        scheme="silx",
    )

    assert find_histogram(
        volume=volume,
        scan=EDFTomoScan(scan=str(tmp_path)),
    ) == DataUrl(
        file_path=histogram_file,
        data_path="entry/histogram/results/data",
        scheme="silx",
    )

    assert find_histogram(
        volume=volume,
        scan=NXtomoScan(scan=str(tmp_path), entry="entry"),
    ) == DataUrl(
        file_path=histogram_file,
        data_path="entry/histogram/results/data",
        scheme="silx",
    )


@pytest.mark.parametrize("input_dtype", (numpy.float32, numpy.float64, numpy.uint8, numpy.uint16))
def test_clamp_and_rescale_data(input_dtype):
    """
    test 'rescale_data' function
    """
    array = numpy.linspace(
        start=1,
        stop=100,
        num=100,
        endpoint=True,
        dtype=input_dtype,
    ).reshape(10, 10)

    rescaled_array = clamp_and_rescale_data(
        data=array,
        new_min=10,
        new_max=90,
        rescale_min_percentile=20,  # provided to insure they will be ignored
        rescale_max_percentile=80,  # provided to insure they will be ignored
    )
    assert rescaled_array.min() == 10
    assert rescaled_array.max() == 90
    numpy.testing.assert_equal(rescaled_array.flatten()[0:10], numpy.array([10] * 10))
    numpy.testing.assert_equal(rescaled_array.flatten()[90:100], numpy.array([90] * 10))


def test_cast_volume(tmp_path):
    """
    test cast_volume
    """
    raw_data = numpy.linspace(
        start=1,
        stop=100,
        num=100,
        endpoint=True,
        dtype=numpy.float64,
    ).reshape(1, 10, 10)
    volume_hdf5_file_path = os.path.join(tmp_path, "myvolume.hdf5")
    volume_hdf5 = HDF5Volume(
        file_path=volume_hdf5_file_path,
        data_path="myentry",
        data=raw_data,
    )
    volume_edf = EDFVolume(
        folder=os.path.join(tmp_path, "volume_folder"),
    )

    # test when no histogram existing
    cast_volume(
        input_volume=volume_hdf5,
        output_volume=volume_edf,
        output_data_type=numpy.dtype(numpy.uint16),
        rescale_min_percentile=10,
        rescale_max_percentile=90,
        save=True,
        store=True,
    )
    # if percentiles 10 and 90 provided, no data_min and data_max then they will be computed from data min / max

    # append histogram
    with h5py.File(volume_hdf5_file_path, mode="a") as h5s:
        hist = numpy.array([20, 20, 20, 20, 20, 20])
        bins = numpy.array([0, 20, 40, 60, 80, 100])
        h5s["myentry/histogram/results/data"] = numpy.vstack((hist, bins))

    # and test it again
    volume_edf.overwrite = True
    cast_volume(
        input_volume=volume_hdf5,
        output_volume=volume_edf,
        output_data_type=numpy.dtype(numpy.uint16),
        rescale_min_percentile=20,
        rescale_max_percentile=60,
        save=True,
        store=True,
    )

    # test to cast the already cast volumes
    volume_tif = EDFVolume(
        folder=os.path.join(tmp_path, "second_volume_folder"),
    )
    volume_tif.overwrite = True
    cast_volume(
        input_volume=volume_edf,
        output_volume=volume_tif,
        output_data_type=numpy.dtype(numpy.uint8),
        save=True,
        store=True,
    )
    assert volume_tif.data.dtype == numpy.uint8

    volume_tif.overwrite = False
    with pytest.raises(OSError):
        cast_volume(
            input_volume=volume_edf,
            output_volume=volume_tif,
            output_data_type=numpy.dtype(numpy.uint8),
            save=True,
            store=True,
        )


@pytest.mark.skipif(not __have_jp2k__, reason="need jp2k (glymur) for this test")
def test_jp2k_compression_ratios(tmp_path):
    """
    simple test to make sure the compression ratios are handled
    """
    import glymur

    raw_data = numpy.random.random(
        size=(1, 2048, 2048),
    )
    raw_data *= 2048.000005
    volume_hdf5_file_path = os.path.join(tmp_path, "myvolume.hdf5")
    volume_hdf5 = HDF5Volume(
        file_path=volume_hdf5_file_path,
        data_path="myentry",
        data=raw_data,
    )

    volume_jp2k_ratios_0 = JP2KVolume(
        folder=os.path.join(tmp_path, "volume_folder"),
        cratios=(100, 10),
    )

    volume_jp2k_ratios_1 = JP2KVolume(
        folder=os.path.join(tmp_path, "volume_folder_2"),
        cratios=(1000, 100),
    )

    # test when no histogram existing
    cast_volume(
        input_volume=volume_hdf5,
        output_volume=volume_jp2k_ratios_0,
        output_data_type=numpy.dtype(numpy.uint16),
        save=True,
        store=True,
    )

    cast_volume(
        input_volume=volume_hdf5,
        output_volume=volume_jp2k_ratios_1,
        output_data_type=numpy.dtype(numpy.uint16),
        save=True,
        store=True,
    )

    # make sure the ratio have been taking into account
    frame_0 = glymur.Jp2k(next(volume_jp2k_ratios_0.browse_data_files()))
    frame_0.layer = 0

    frame_1 = glymur.Jp2k(next(volume_jp2k_ratios_1.browse_data_files()))
    frame_1.layer = 0

    assert not numpy.array_equal(frame_0, frame_1)



from os import remove
import numpy as np
from .. import version
from .reconstruct import get_reconstructor
from .cli_configs import MultiCorConfig
from .utils import parse_params_values
from ..utils import view_as_images_stack


def get_user_cors(cors):
    """
    From a user-provided str describing the centers of rotation, build a list.
    """
    cors = cors.strip("[()]")
    cors = cors.split(",")
    cors = [c.strip() for c in cors]
    cors_list = []
    for c in cors:
        if ":" in c:
            if c.count(":") != 2:
                raise ValueError("Malformed range format for '%s': expected format start:stop:step" % c)
            start, stop, step = c.split(":")
            c_list = np.arange(float(start), float(stop), float(step)).tolist()
        else:
            c_list = [float(c)]
        cors_list.extend(c_list)
    return cors_list


def main():
    args = parse_params_values(
        MultiCorConfig,
        parser_description=f"Perform a tomographic reconstruction of a single slice using multiple centers of rotation",
        program_version="nabu " + version,
    )

    reconstructor = get_reconstructor(
        args,
        # Put a dummy CoR to avoid crash in both full-FoV and extended-FoV.
        # It will be overwritten later by the user-defined CoRs
        overwrite_options={"reconstruction/rotation_axis_position": 10.0},
    )

    if reconstructor.delta_z > 1:
        raise ValueError("Only slice reconstruction can be used (have delta_z = %d)" % reconstructor.delta_z)

    reconstructor.reconstruct()  # warm-up, spawn pipeline

    pipeline = reconstructor.pipeline
    file_prefix = pipeline.processing_options["save"]["file_prefix"]
    #####
    # Remove the first reconstructed file (not used here)
    last_file = list(pipeline.writer.writer.browse_data_files())[-1]
    try:
        remove(last_file)
    except:
        pass
    ######

    cors = get_user_cors(args["cor"])

    all_recs = []
    rec_instance = pipeline.reconstruction

    for cor in cors:
        # Re-configure with new CoR
        pipeline.processing_options["reconstruction"]["rotation_axis_position"] = cor
        pipeline.processing_options["save"]["file_prefix"] = file_prefix + "_%.03f" % cor
        pipeline._init_writer(create_subfolder=False, single_output_file_initialized=False)

        # Get sinogram into contiguous array
        # TODO Can't do memcpy2D ?! It used to work in cuda 11.
        # For now: transfer to host... not optimal
        sino = pipeline._d_radios[:, pipeline._d_radios.shape[1] // 2, :].get()  # pylint: disable=E1136

        if pipeline.process_config.do_halftomo:
            # re-initialize FBP object, because in half-tomography the output slice size is a function of CoR
            options = pipeline.processing_options["reconstruction"]
            rec_instance = pipeline.FBPClass(
                sino.shape,
                angles=options["angles"],
                rot_center=cor,
                filter_name=options["fbp_filter_type"] or "none",
                halftomo=options["enable_halftomo"],
                # slice_roi=self.process_config.rec_roi,
                padding_mode=options["padding_type"],
                extra_options={
                    "scale_factor": 1.0 / options["voxel_size_cm"][0],
                    "axis_correction": options["axis_correction"],
                    "centered_axis": options["centered_axis"],
                    "clip_outer_circle": options["clip_outer_circle"],
                    "filter_cutoff": options["fbp_filter_cutoff"],
                },
            )
        else:
            pipeline.reconstruction.reset_rot_center(cor)

        # Run reconstruction
        rec = rec_instance.fbp(sino)
        # if return_all_recs:
        # all_recs.append(rec)
        rec_3D = view_as_images_stack(rec)  # writer wants 3D data

        # Write
        pipeline.writer.write_data(rec_3D)
        reconstructor.logger.info("Wrote %s" % pipeline.writer.fname)

    return 0


if __name__ == "__main__":
    main()



from ..fullfield.dataset_validator import *
from ...utils import copy_dict_items


class HelicalDatasetValidator(FullFieldDatasetValidator):
    """Allows more freedom in the choice of the slice indices"""

    # this in the fullfield base  class is instead True
    _check_also_z = False

    def _check_slice_indices(self):
        """Slice indices can be far beyond what fullfield pipeline accepts, no check here, but
        Nabu expects that rec_region is initialised here"""

        what = ["start_x", "end_x", "start_y", "end_y", "start_z", "end_z"]

        self.rec_region = copy_dict_items(self.rec_params, what)

        return



from itertools import product
import numpy as np
import pytest
from nabu.misc.binning import *


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.data = np.arange(100 * 99, dtype=np.uint16).reshape(100, 99)
    cls.tol = 1e-5


@pytest.mark.usefixtures("bootstrap")
class TestBinning:
    def testBinning(self):
        """
        Test the general-purpose binning function with an image defined by its indices.
        The test "image" is an array where entry [i, j] equals i * Nx + j (where Nx = array.shape[1]).
        Let (b_i, b_j) be the binning factor along each dimension,
        then the binned array at position [p_i, p_j] is equal to
           1/(b_i*b_j) * Sum(Sum(Nx*i + j, (j, p_j, p_j+b_j-1)), (i, p_i, p_i+b_i-1))
        which happens to be equal to
           (Nx * b_i + 2*Nx * p_i - Nx + b_j + 2*p_j - 1) /2
        """

        def get_reference_binned_image(img_shape, bin_factor):
            # reference[p_i, p_j] = 0.5 * (Nx * b_i + 2*Nx * p_i - Nx + b_j + 2*p_j - 1)
            Ny, Nx = img_shape
            img_shape_reduced = tuple(s - (s % b) for s, b in zip(img_shape, bin_factor))
            b_i, b_j = bin_factor
            inds_i, inds_j = np.indices(img_shape_reduced)
            p_i = inds_i[::b0, ::b1]
            p_j = inds_j[::b0, ::b1]
            return 0.5 * (Nx * b_i + 2 * Nx * p_i - Nx + b_j + 2 * p_j - 1)

        # Various test settings
        binning_factors = [2, 3, 4, 5, 6, 8, 10]
        n_items = [63, 64, 65, 66, 125, 128, 130]

        # Yep, that's 2401 tests...
        params = product(n_items, n_items, binning_factors, binning_factors)
        for s0, s1, b0, b1 in params:
            img = np.arange(s0 * s1).reshape((s0, s1))
            ref = get_reference_binned_image(img.shape, (b0, b1))
            res = binning(img, (b0, b1))

            assert np.allclose(res, ref)



import math
import numpy as np
from ..utils import deprecated_class, deprecation_warning, is_scalar
from ..misc import fourier_filters
from .alignment import AlignmentBase, plt, progress_bar, local_fftn, local_ifftn


# three possible  values for the validity check, which can optionally be returned by the find_shifts methods
cor_result_validity = {
    "unknown": "unknown",
    "sound": "sound",
    "correct": "sound",
    "questionable": "questionable",
}


class CenterOfRotation(AlignmentBase):

    def find_shift(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        side=None,
        shift_axis: int = -1,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
        return_validity=False,
        return_relative_to_middle=None,
    ):
        """Find the Center of Rotation (CoR), given two images.

        This method finds the half-shift between two opposite images, by
        means of correlation computed in Fourier space.

        The output of this function, allows to compute motor movements for
        aligning the sample rotation axis. Given the following values:

           - L1: distance from source to motor
           - L2: distance from source to detector
           - ps: physical pixel size
           - v: output of this function

        displacement of motor = (L1 / L2 * ps) * v

        Parameters
        ----------
        img_1: numpy.ndarray
            First image
        img_2: numpy.ndarray
            Second image, it needs to have been flipped already (e.g. using numpy.fliplr).
        shift_axis: int
            Axis along which we want the shift to be computed. Default is -1 (horizontal).
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`
        return_validity: a boolean, defaults to false
            if set to True adds a second return value which may have three string values.
            These values are "unknown", "sound", "questionable".
            It will be "uknown" if the  validation method is not implemented
            and it will be "sound" or "questionable" if it is implemented.


        Raises
        ------
        ValueError
            In case images are not 2-dimensional or have different sizes.

        Returns
        -------
        float
            Estimated center of rotation position from the center of the RoI in pixels.

        Examples
        --------
        The following code computes the center of rotation position for two
        given images in a tomography scan, where the second image is taken at
        180 degrees from the first.

        >>> radio1 = data[0, :, :]
        ... radio2 = np.fliplr(data[1, :, :])
        ... CoR_calc = CenterOfRotation()
        ... cor_position = CoR_calc.find_shift(radio1, radio2)

        Or for noisy images:

        >>> cor_position = CoR_calc.find_shift(radio1, radio2, median_filt_shape=(3, 3))
        """
        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotation.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---

        self._check_img_pair_sizes(img_1, img_2)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        cc = self._compute_correlation_fft(img_1, img_2, padding_mode, high_pass=high_pass, low_pass=low_pass)
        img_shape = img_2.shape
        cc_vs = np.fft.fftfreq(img_shape[-2], 1 / img_shape[-2])
        cc_hs = np.fft.fftfreq(img_shape[-1], 1 / img_shape[-1])

        (f_vals, fv, fh) = self.extract_peak_region_2d(cc, peak_radius=peak_fit_radius, cc_vs=cc_vs, cc_hs=cc_hs)
        fitted_shifts_vh = self.refine_max_position_2d(f_vals, fv, fh)

        estimated_cor = fitted_shifts_vh[shift_axis] / 2.0

        if is_scalar(side):
            near_pos = side - (img_1.shape[-1] - 1) / 2
            if (
                np.abs(near_pos - estimated_cor) / near_pos > 0.2
            ):  # For comparison, near_pos is RELATIVE to the middle of image (as estimated_cor is).
                validity_check_result = cor_result_validity["questionable"]
            else:
                validity_check_result = cor_result_validity["sound"]
        else:
            validity_check_result = cor_result_validity["unknown"]

        if not (return_relative_to_middle):
            estimated_cor += (img_1.shape[-1] - 1) / 2

        if return_validity:
            return estimated_cor, validity_check_result
        else:
            return estimated_cor


class CenterOfRotationSlidingWindow(CenterOfRotation):
    def find_shift(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        side="center",
        window_width=None,
        roi_yxhw=None,
        median_filt_shape=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
        return_validity=False,
        return_relative_to_middle=None,
    ):
        """Semi-automatically find the Center of Rotation (CoR), given two images
        or sinograms. Suitable for half-aquisition scan.

        This method finds the half-shift between two opposite images,  by
        minimizing difference over a moving window.

        Parameters and usage is the same as CenterOfRotation, except for the following two parameters.

        Parameters
        ----------
        side: string or float, optional
            Expected region of the CoR. Allowed values: 'left', 'center' or 'right'. Default is 'center'
        window_width: int, optional
            Width of window that will slide on the other image / part of the sinogram. Default is None.
        """
        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotationSlidingWindow.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---
        validity_check_result = cor_result_validity["unknown"]

        if side is None:
            raise ValueError("Side should be one of 'left', 'right', 'center' or a scalar. 'None' was given instead")

        self._check_img_pair_sizes(img_1, img_2)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(
            img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, high_pass=high_pass, low_pass=low_pass
        )
        img_2 = self._prepare_image(
            img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, high_pass=high_pass, low_pass=low_pass
        )
        img_shape = img_2.shape
        img_width = img_shape[-1]

        if isinstance(side, str):
            if window_width is None:
                if side == "center":
                    window_width = round(img_width / 4.0 * 3.0)
                else:
                    window_width = round(img_width / 10)
            window_shift = window_width // 2
            window_width = window_shift * 2 + 1
            if side == "right":
                win_2_start = 0
            elif side == "left":
                win_2_start = img_width - window_width
            else:
                win_2_start = img_width // 2 - window_shift
        else:
            abs_pos = int(side + img_width // 2)
            window_fraction = 0.1  # Hard-coded ?

            window_width = round(window_fraction * img_width)
            window_shift = window_width // 2
            window_width = window_shift * 2 + 1

            win_2_start = np.clip(abs_pos - window_shift, 0, img_width // 2 - 1)
            win_2_start = img_width // 2 - 1 - win_2_start

        win_1_start_seed = 0
        # number of pixels where the window will "slide".
        n = img_width - window_width

        win_2_end = win_2_start + window_width

        diffs_mean = np.zeros((n,), dtype=img_1.dtype)
        diffs_std = np.zeros((n,), dtype=img_1.dtype)

        for ii in progress_bar(range(n), verbose=self.verbose):
            win_1_start = win_1_start_seed + ii
            win_1_end = win_1_start + window_width
            img_diff = img_1[:, win_1_start:win_1_end] - img_2[:, win_2_start:win_2_end]
            diffs_abs = np.abs(img_diff)
            diffs_mean[ii] = diffs_abs.mean()
            diffs_std[ii] = diffs_abs.std()

        diffs_mean = diffs_mean.min() - diffs_mean
        win_ind_max = np.argmax(diffs_mean)

        diffs_std = diffs_std.min() - diffs_std
        if not win_ind_max == np.argmax(diffs_std):
            self.logger.warning(
                "Minimum mean difference and minimum std-dev of differences do not coincide. "
                + "This means that the validity of the found solution might be questionable."
            )
            validity_check_result = cor_result_validity["questionable"]
        else:
            validity_check_result = cor_result_validity["sound"]

        (f_vals, f_pos) = self.extract_peak_regions_1d(diffs_mean, peak_radius=peak_fit_radius)
        win_pos_max, win_val_max = self.refine_max_position_1d(f_vals, return_vertex_val=True)

        # Derive the COR
        if is_scalar(side):
            cor_h = -(win_2_start - (win_1_start_seed + win_ind_max + win_pos_max)) / 2.0
            cor_pos = -(win_2_start - (win_1_start_seed + np.arange(n))) / 2.0
        else:
            cor_h = -(win_2_start - (win_ind_max + win_pos_max)) / 2.0
            cor_pos = -(win_2_start - np.arange(n)) / 2.0

        if (side == "right" and win_ind_max == 0) or (side == "left" and win_ind_max == n):
            self.logger.warning("Sliding window width %d might be too large!" % window_width)

        if self.verbose:
            print("Lowest difference window: index=%d, range=[0, %d]" % (win_ind_max, n))
            print("CoR tested for='%s', found at voxel=%g (from center)" % (side, cor_h))

            f, ax = plt.subplots(1, 1)
            self._add_plot_window(f, ax=ax)
            ax.stem(cor_pos, diffs_mean, label="Mean difference")
            ax.stem(cor_h, win_val_max, linefmt="C1-", markerfmt="C1o", label="Best mean difference")
            ax.stem(cor_pos, -diffs_std, linefmt="C2-", markerfmt="C2o", label="Std-dev difference")
            ax.set_title("Window dispersions")
            plt.legend()
            plt.show(block=False)

        if not (return_relative_to_middle):
            cor_h += (img_width - 1) / 2.0

        if return_validity:
            return cor_h, validity_check_result
        else:
            return cor_h


class CenterOfRotationGrowingWindow(CenterOfRotation):
    def find_shift(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        side="all",
        min_window_width=11,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
        return_validity=False,
        return_relative_to_middle=None,
    ):
        """Automatically find the Center of Rotation (CoR), given two images or
        sinograms. Suitable for half-aquisition scan.

        This method finds the half-shift between two opposite images,  by
        minimizing difference over a moving window.

        Usage and parameters are the same as CenterOfRotationSlidingWindow, except for the following parameter.

        Parameters
        ----------
        min_window_width: int, optional
            Minimum window width that covers the common region of the two images /
            sinograms. Default is 11.
        """
        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotationGrowingWindow.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---
        validity_check_result = cor_result_validity["unknown"]

        self._check_img_pair_sizes(img_1, img_2)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(
            img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, high_pass=high_pass, low_pass=low_pass
        )
        img_2 = self._prepare_image(
            img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, high_pass=high_pass, low_pass=low_pass
        )
        img_shape = img_2.shape

        def window_bounds(mid_point, window_max_width=img_shape[-1]):
            return (
                np.fmax(np.ceil(mid_point - window_max_width / 2), 0).astype(np.intp),
                np.fmin(np.ceil(mid_point + window_max_width / 2), img_shape[-1]).astype(np.intp),
            )

        img_lower_half_size = np.floor(img_shape[-1] / 2).astype(np.intp)
        img_upper_half_size = np.ceil(img_shape[-1] / 2).astype(np.intp)

        if is_scalar(side):
            self.logger.error(
                "Passing a first CoR guess is not supported for CenterOfRotationGrowingWindow. Using side='all'."
            )
            side = "all"
        if side.lower() == "right":
            win_1_mid_start = img_lower_half_size
            win_1_mid_end = np.floor(img_shape[-1] * 3 / 2).astype(np.intp) - min_window_width
            win_2_mid_start = -img_upper_half_size + min_window_width
            win_2_mid_end = img_upper_half_size
        elif side.lower() == "left":
            win_1_mid_start = -img_lower_half_size + min_window_width
            win_1_mid_end = img_lower_half_size
            win_2_mid_start = img_upper_half_size
            win_2_mid_end = np.ceil(img_shape[-1] * 3 / 2).astype(np.intp) - min_window_width
        elif side.lower() == "center":
            win_1_mid_start = 0
            win_1_mid_end = img_shape[-1]
            win_2_mid_start = 0
            win_2_mid_end = img_shape[-1]
        elif side.lower() == "all":
            win_1_mid_start = -img_lower_half_size + min_window_width
            win_1_mid_end = np.floor(img_shape[-1] * 3 / 2).astype(np.intp) - min_window_width
            win_2_mid_start = -img_upper_half_size + min_window_width
            win_2_mid_end = np.ceil(img_shape[-1] * 3 / 2).astype(np.intp) - min_window_width
        else:
            raise ValueError(
                "Side should be one of 'left', 'right', or 'center' or 'all'. '%s' was given instead" % side.lower()
            )

        n1 = win_1_mid_end - win_1_mid_start
        n2 = win_2_mid_end - win_2_mid_start

        if not n1 == n2:
            raise ValueError(
                "Internal error: the number of window steps for the two images should be the same."
                + "Found the following configuration instead => Side: %s, #1: %d, #2: %d" % (side, n1, n2)
            )

        diffs_mean = np.zeros((n1,), dtype=img_1.dtype)
        diffs_std = np.zeros((n1,), dtype=img_1.dtype)

        for ii in progress_bar(range(n1), verbose=self.verbose):
            win_1 = window_bounds(win_1_mid_start + ii)
            win_2 = window_bounds(win_2_mid_end - ii)
            img_diff = img_1[:, win_1[0] : win_1[1]] - img_2[:, win_2[0] : win_2[1]]
            diffs_abs = np.abs(img_diff)
            diffs_mean[ii] = diffs_abs.mean()
            diffs_std[ii] = diffs_abs.std()

        diffs_mean = diffs_mean.min() - diffs_mean
        win_ind_max = np.argmax(diffs_mean)

        diffs_std = diffs_std.min() - diffs_std
        if not win_ind_max == np.argmax(diffs_std):
            self.logger.warning(
                "Minimum mean difference and minimum std-dev of differences do not coincide. "
                + "This means that the validity of the found solution might be questionable."
            )
            validity_check_result = cor_result_validity["questionable"]
        else:
            validity_check_result = cor_result_validity["sound"]

        (f_vals, f_pos) = self.extract_peak_regions_1d(diffs_mean, peak_radius=peak_fit_radius)
        win_pos_max, win_val_max = self.refine_max_position_1d(f_vals, return_vertex_val=True)

        cor_h = (win_1_mid_start + (win_ind_max + win_pos_max) - img_upper_half_size) / 2.0

        if (side.lower() == "right" and win_ind_max == 0) or (side.lower() == "left" and win_ind_max == n1):
            self.logger.warning("Minimum growing window width %d might be too large!" % min_window_width)

        if self.verbose:
            cor_pos = (win_1_mid_start + np.arange(n1) - img_upper_half_size) / 2.0

            self.logger.info("Lowest difference window: index=%d, range=[0, %d]" % (win_ind_max, n1))
            self.logger.info("CoR tested for='%s', found at voxel=%g (from center)" % (side, cor_h))

            f, ax = plt.subplots(1, 1)
            self._add_plot_window(f, ax=ax)
            ax.stem(cor_pos, diffs_mean, label="Mean difference")
            ax.stem(cor_h, win_val_max, linefmt="C1-", markerfmt="C1o", label="Best mean difference")
            ax.stem(cor_pos, -diffs_std, linefmt="C2-", markerfmt="C2o", label="Std-dev difference")
            ax.set_title("Window dispersions")
            plt.show(block=False)

        if not (return_relative_to_middle):
            cor_h += (img_shape[-1] - 1) / 2.0

        if return_validity:
            return cor_h, validity_check_result
        else:
            return cor_h


class CenterOfRotationAdaptiveSearch(CenterOfRotation):
    """This adaptive method works by applying a gaussian which highlights, by apodisation, a region
    which can possibly contain the good center of rotation.
    The whole image is spanned during several applications of the apodisation. At each application
    the apodisation function, which is a gaussian, is moved to a new guess position.
    The lenght of the step, by which the gaussian is moved, and its sigma are
    obtained by multiplying the shortest distance from the left or right border with
    a self.step_fraction and  self.sigma_fraction factors which ensure global overlapping.
    for each step a region around the CoR  of each image is selected, and the regions of the two images
    are compared to  calculate a cost function. The value of the cost function, at its minimum
    is used to select the best step at which the CoR is taken as final result.
    The option filtered_cost= True (default) triggers the filtering (according to low_pass and high_pass)
    of the two images which are used for he cost function. ( Note: the low_pass and high_pass options
    are used, if given, also without the filtered_cost option, by being passed to the base class
    CenterOfRotation )
    """

    sigma_fraction = 1.0 / 4.0
    step_fraction = 1.0 / 6.0

    def find_shift(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        high_pass=None,
        low_pass=None,
        margins=None,
        filtered_cost=True,
        return_validity=False,
        return_relative_to_middle=None,
    ):
        """Find the Center of Rotation (CoR), given two images.

        This method finds the half-shift between two opposite images, by
        means of correlation computed in Fourier space.
        A global search is done on on the detector span (minus a margin) without assuming centered scan conditions.

        Usage and parameters are the same as CenterOfRotation, except for the following parameters.

        Parameters
        ----------
        margins:  None or a couple of floats or ints
            if margins is None or in the form of  (margin1,margin2) the search is done between margin1 and  dim_x-1-margin2.
            If left to None then by default (margin1,margin2)  = ( 10, 10 ).
        filtered_cost: boolean.
            True by default. It triggers the use of filtered images in the calculation of the cost function.
        """
        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotationAdaptiveSearch.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---
        validity_check_result = cor_result_validity["unknown"]

        self._check_img_pair_sizes(img_1, img_2)

        used_type = img_1.dtype

        roi_yxhw = self._determine_roi(img_1.shape, roi_yxhw)

        if filtered_cost and (low_pass is not None or high_pass is not None):
            img_filter = fourier_filters.get_bandpass_filter(
                img_1.shape[-2:],
                cutoff_lowpass=low_pass,
                cutoff_highpass=high_pass,
                use_rfft=True,
                data_type=self.data_type,
            )
            # fft2 and iff2 use axes=(-2, -1) by default
            img_filtered_1 = local_ifftn(local_fftn(img_1, axes=(-2, -1)) * img_filter, axes=(-2, -1)).real
            img_filtered_2 = local_ifftn(local_fftn(img_2, axes=(-2, -1)) * img_filter, axes=(-2, -1)).real
        else:
            img_filtered_1 = img_1
            img_filtered_2 = img_2

        img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        img_filtered_1 = self._prepare_image(img_filtered_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_filtered_2 = self._prepare_image(img_filtered_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        dim_radio = img_1.shape[1]

        if margins is None:
            lim_1, lim_2 = 10, dim_radio - 1 - 10
        else:
            lim_1, lim_2 = margins
            lim_2 = dim_radio - 1 - lim_2

        if lim_1 < 1:
            lim_1 = 1
        if lim_2 > dim_radio - 2:
            lim_2 = dim_radio - 2

        if lim_2 <= lim_1:
            message = (
                "Image shape or cropped selection too small for global search."
                + " After removal of the margins the search limits collide."
                + " The cropped size is %d\n" % (dim_radio)
            )
            raise ValueError(message)

        found_centers = []
        x_cor = lim_1
        while x_cor < lim_2:
            tmp_sigma = (
                min(
                    (img_1.shape[1] - x_cor),
                    (x_cor),
                )
                * self.sigma_fraction
            )

            tmp_x = (np.arange(img_1.shape[1]) - x_cor) / tmp_sigma
            apodis = np.exp(-tmp_x * tmp_x / 2.0)

            x_cor_rel = x_cor - (img_1.shape[1] // 2)

            img_1_apodised = img_1 * apodis

            try:
                cor_position = CenterOfRotation.find_shift(
                    self,
                    img_1_apodised.astype(used_type),
                    img_2.astype(used_type),
                    low_pass=low_pass,
                    high_pass=high_pass,
                    roi_yxhw=roi_yxhw,
                    return_relative_to_middle=True,
                )
            except ValueError as err:
                if "positions are outside the input margins" in str(err):
                    x_cor = min(x_cor + x_cor * self.step_fraction, x_cor + (dim_radio - x_cor) * self.step_fraction)
                    continue
            except:
                message = "Unexpected error from base class CenterOfRotation.find_shift in CenterOfRotationAdaptiveSearch.find_shift  : {err}".format(
                    err=err
                )
                self.logger.error(message)
                raise

            p_1 = cor_position * 2
            if cor_position < 0:
                p_2 = img_2.shape[1] + cor_position * 2
            else:
                p_2 = -img_2.shape[1] + cor_position * 2

            if abs(x_cor_rel - p_1 / 2) < abs(x_cor_rel - p_2 / 2):
                cor_position = p_1 / 2
            else:
                cor_position = p_2 / 2

            cor_in_img = img_1.shape[1] // 2 + cor_position
            tmp_sigma = (
                min(
                    (img_1.shape[1] - cor_in_img),
                    (cor_in_img),
                )
                * self.sigma_fraction
            )

            M1 = int(round(cor_position + img_1.shape[1] // 2)) - int(round(tmp_sigma))
            M2 = int(round(cor_position + img_1.shape[1] // 2)) + int(round(tmp_sigma))

            piece_1 = img_filtered_1[:, M1:M2]
            piece_2 = img_filtered_2[:, img_1.shape[1] - M2 : img_1.shape[1] - M1]

            if piece_1.size and piece_2.size:
                piece_1 = piece_1 - piece_1.mean()
                piece_2 = piece_2 - piece_2.mean()
                energy = np.array(piece_1 * piece_1 + piece_2 * piece_2, "d").sum()
                diff_energy = np.array((piece_1 - piece_2) * (piece_1 - piece_2), "d").sum()
                cost = diff_energy / energy

                if not np.isnan(cost):
                    if tmp_sigma * 2 > abs(x_cor_rel - cor_position):
                        found_centers.append([cost, abs(x_cor_rel - cor_position), cor_position, energy])

            x_cor = min(x_cor + x_cor * self.step_fraction, x_cor + (dim_radio - x_cor) * self.step_fraction)

        if len(found_centers) == 0:
            message = "Unable to find any valid CoR candidate in {my_class}.find_shift ".format(
                my_class=self.__class__.__name__
            )
            raise ValueError(message)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Now build the neigborhood of the minimum as a list of five elements:
        # the minimum in the middle of the two before, and the two after

        filtered_found_centers = []
        for i in range(len(found_centers)):
            if i > 0:
                if abs(found_centers[i][2] - found_centers[i - 1][2]) < 0.5:
                    filtered_found_centers.append(found_centers[i])
                    continue
            if i + 1 < len(found_centers):
                if abs(found_centers[i][2] - found_centers[i + 1][2]) < 0.5:
                    filtered_found_centers.append(found_centers[i])
                    continue

        if len(filtered_found_centers):
            found_centers = filtered_found_centers

        min_choice = min(found_centers)
        index_min_choice = found_centers.index(min_choice)
        min_neighborood = [
            found_centers[i][2] if (i >= 0 and i < len(found_centers)) else math.nan
            for i in range(index_min_choice - 2, index_min_choice + 2 + 1)
        ]

        score_right = 0
        for i_pos in [3, 4]:
            if abs(min_neighborood[i_pos] - min_neighborood[2]) < 0.5:
                score_right += 1
            else:
                break

        score_left = 0
        for i_pos in [1, 0]:
            if abs(min_neighborood[i_pos] - min_neighborood[2]) < 0.5:
                score_left += 1
            else:
                break

        if score_left + score_right >= 2:
            validity_check_result = cor_result_validity["sound"]
        else:
            self.logger.warning(
                "Minimum mean difference and minimum std-dev of differences do not coincide. "
                + "This means that the validity of the found solution might be questionable."
            )
            validity_check_result = cor_result_validity["questionable"]

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # An informative message in case one wish to look at how it has gone
        informative_message = " ".join(
            ["CenterOfRotationAdaptiveSearch found this neighborood of the optimal position:"]
            + [str(t) if not math.isnan(t) else "N.A." for t in min_neighborood]
        )
        self.logger.debug(informative_message)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # The return value is the optimum which had been placed in the middle of the neighborood
        cor_position = min_neighborood[2]

        if not (return_relative_to_middle):
            cor_position += (img_1.shape[-1] - 1) / 2.0

        if return_validity:
            return cor_position, validity_check_result
        else:
            return cor_position

    __call__ = find_shift


class CenterOfRotationOctaveAccurate(CenterOfRotation):
    """This is a Python implementation of Octave/fastomo3/accurate COR estimator.
    The Octave 'accurate' function is renamed `local_correlation`.
    The Nabu standard `find_shift` has the same API as the other COR estimators (sliding, growing...)

    """

    def _cut(self, im, nrows, ncols, new_center_row=None, new_center_col=None):
        """Cuts a sub-matrix out of a larger matrix.
        Cuts in the center of the original matrix, except if new center is specified
        NO CHECKING of validity indices sub-matrix!

        Parameters
        ----------
        im : array.
            Original matrix
        nrows : int
            Number of rows in the output matrix.
        ncols : int
            Number of columns in the output matrix.
        new_center_row : int
            Index of center row around which to cut (default: None, i.e. center)
        new_center_col : int
            Index of center column around which to cut (default: None, i.e. center)

        Returns
        -------
        nrows x ncols array.

        Examples
        --------
        im_roi = cut(im, 1024, 1024)                -> cut center 1024x1024 pixels
        im_roi = cut(im, 1024, 1024, 600.5, 700.5)  -> cut 1024x1024 pixels around pixels (600-601, 700-701)

        Author: P. Cloetens <cloetens@esrf.eu>
        2023-11-06 J. Lesaint <jerome.lesaint@esrf.fr>

        * See octave-archive for the original Octave code.
        * 2023-11-06: Python implementation. Comparison seems OK.
        """
        [n, m] = im.shape
        if new_center_row is None:
            new_center_row = (n + 1) / 2
        if new_center_col is None:
            new_center_col = (m + 1) / 2

        rb = int(np.round(0.5 + new_center_row - nrows / 2))
        rb = int(np.round(new_center_row - nrows / 2))
        re = int(nrows + rb)
        cb = int(np.round(0.5 + new_center_col - ncols / 2))
        cb = int(np.round(new_center_col - ncols / 2))
        ce = int(ncols + cb)

        return im[rb:re, cb:ce]

    def _checkifpart(self, rapp, rapp_hist):
        res = 0
        for k in range(rapp_hist.shape[0]):
            if np.allclose(rapp, rapp_hist[k, :]):
                res = 1
                return res
        return res

    def _interpolate(self, input, shift, mode="mean", interpolation_method="linear"):
        """Applies to the input a translation by a vector `shift`. Based on
        `scipy.ndimage.affine_transform` function.
        JL: This Octave function was initially used in the refine clause of the local_correlation (Octave find_shift).
        Since find_shift is always called with refine=False in Octave, refine is not implemented (see local_interpolation())
        and this function becomes useless.

        Parameters
        ----------
        input : array
            Array to which the translation is applied.
        shift : tuple, list or array of length 2.
        mode : str
            Type of padding applied to the unapplicable areas of the output image.
            Default `mean` is a constant padding with the mean of the input array.
            `mode` must belong to 'reflect', 'grid-mirror', 'constant', 'grid-constant', 'nearest', 'mirror', 'grid-wrap', 'wrap'
            See `scipy.ndimage.affine_transform` for details.
        interpolation_method : str or int.
            The interpolation is based on spline interpolation.
            Either 0, 1, 2, 3, 4 or 5: order of the spline interpolation functions.
            Or one among 'linear','cubic','pchip','nearest','spline' (Octave legacy).
                'nearest' is equivalent to 0
                'linear' is equivalent to 1
                'cubic','pchip','spline' are equivalent to 3.
        """
        admissible_modes = (
            "reflect",
            "grid-mirror",
            "constant",
            "grid-constant",
            "nearest",
            "mirror",
            "grid-wrap",
            "wrap",
        )
        admissible_interpolation_methods = ("linear", "cubic", "pchip", "nearest", "spline")

        from scipy.ndimage import affine_transform

        [s0, s1] = shift
        matrix = np.zeros([2, 3], dtype=float)
        matrix[0, 0] = 1.0
        matrix[1, 1] = 1.0
        matrix[:, 2] = [-s0, -s1]  # JL: due to transf. convention diff in Octave and scipy (push fwd vs pull back)

        if interpolation_method == "nearest":
            order = 0
        elif interpolation_method == "linear":
            order = 1
        elif interpolation_method in ("pchip", "cubic", "spline"):
            order = 3
        elif interpolation_method in (0, 1, 2, 3, 4, 5):
            order = interpolation_method
        else:
            raise ValueError(
                f"Interpolation method is {interpolation_method} and should either an integer between 0 (inc.) and 5 (inc.) or in {admissible_interpolation_methods}."
            )

        if mode == "mean":
            mode = "constant"
            cval = input.mean()
            return affine_transform(input, matrix, mode=mode, order=order, cval=cval)
        elif mode not in admissible_modes:
            raise ValueError(f"Pad method is {mode} and should be in {admissible_modes}.")

        return affine_transform(input, matrix, mode=mode, order=order)

    def _local_correlation(
        self,
        z1,
        z2,
        maxsize=[5, 5],
        cor_estimate=[0, 0],
        refine=None,
        pmcc=False,
        normalize=True,
    ):
        """Returns the 2D shift in pixels between two images.
        It looks for a local optimum around the initial shift cor_estimate
        and within a window 'maxsize'.
        It uses variance of the difference of the normalized images or PMCC
        It adapts the shift estimate in case optimum is at the edge of the window
        If 'maxsize' is set to 0, it will only use approximate shift (+ refine possibly)
        Set 'cor_estimate' to allow for the use of any initial shift estimation.

        When not successful (stuck in loop or edge reached), returns [nan nan]
        Positive values corresponds to moving z2 to higher values of the index
        to compensate drift: interpolate(f)(z2, row, column)

        Parameters
        ----------
        z1,z2 : 2D arrays.
            The two (sub)images to be compared.

        maxsize : 2-list. Default [5,5]
            Size of the search window.

        cor_estimate:
            Initial guess of the center of rotation.

        refine: Boolean or None (default is None)
            Wether the initial guess should be refined of not.

        pmcc: Boolean (default is False)
            Use Pearson correlation coefficient  i.o. variance.

        normalize: Boolean (default is True)
            Set mean of each image to 1 if True.

        Returns
        -------
        c = [row,column] (or [NaN,NaN] if unsuccessful.)

        2007-01-05 P. Cloetens cloetens@esrf.eu
        * Initial revision
        2023-11-10 J. Lesaint jerome.lesaint@esrf.fr
        * Python conversion.
        """

        if type(maxsize) in (float, int):
            maxsize = [int(maxsize), int(maxsize)]
        elif type(maxsize) in (tuple, list):
            maxsize = [int(maxsize[0]), int(maxsize[1])]
        elif maxsize in ([], None, ""):
            maxsize = [5, 5]

        if refine is None:
            refine = np.allclose(maxsize, 0.0)

        if normalize:
            z1 /= np.mean(z1)
            z2 /= np.mean(z2)

        #####################################
        # JL : seems useless since func is always called with a first approximate.
        ## determination of approximative shift (manually or Fourier correlation)
        # if isinstance(cor_estimate,str):
        #    if cor_estimate in ('fft','auto','fourier'):
        #        padding_mode = None
        #        cor_estimate = self._compute_correlation_fft(
        #            z1,
        #            z2,
        #            padding_mode,
        #            high_pass=self.high_pass,
        #            low_pass=self.low_pass
        #        )
        #    elif cor_estimate in ('manual','man','m'):
        #        cor_estimate = None
        #        # No ImageJ plugin here :
        #        # rapp = ij_align(z1,z2)

        ####################################
        # check if refinement with realspace correlation is required
        # otherwise keep result as it is
        if np.allclose(maxsize, 0):
            shiftfound = 1
            if refine:
                c = np.round(np.array(cor_estimate, dtype=int))
            else:
                c = np.array(cor_estimate, dtype=int)
        else:
            shiftfound = 0
            cor_estimate = np.round(np.array(cor_estimate, dtype=int))

        rapp_hist = []
        if np.sum(np.abs(cor_estimate) + 1 >= z1.shape):
            self.logger.debug(f"Approximate shift of [{cor_estimate[0]},{cor_estimate[1]}] is too large, setting [0 0]")
            cor_estimate = np.array([0, 0])
        maxsize = np.minimum(maxsize, np.floor((np.array(z1.shape) - 1) / 2)).astype(int)
        maxsize = np.minimum(maxsize, np.array(z1.shape) - np.abs(cor_estimate) - 1).astype(int)

        while not shiftfound:
            # Set z1 region
            # Rationale: the (shift[0]+maxsize[0]:,shift[1]+maxsize[1]:) block of z1 should match
            # the (maxsize[0]:,maxisze[1]:)-upper-left corner of z2.
            # We first extract this z1 block.
            # Then, take moving z2-block according to maxsize.
            # Of course, care must be taken with borders, hence the various max,min calls.

            # Extract the reference block
            shape_ar = np.array(z1.shape)
            cor_ar = np.array(cor_estimate)
            maxsize_ar = np.array(maxsize)

            z1beg = np.maximum(cor_ar + maxsize_ar, np.zeros(2, dtype=int))
            z1end = shape_ar + np.minimum(cor_ar - maxsize_ar, np.zeros(2, dtype=int))

            z1p = z1[z1beg[0] : z1end[0], z1beg[1] : z1end[1]].flatten()

            # Build local correlations array.
            window_shape = (2 * int(maxsize[0]) + 1, 2 * int(maxsize[1]) + 1)
            cc = np.zeros(window_shape)

            # Prepare second block indices
            z2beg = (cor_ar + maxsize_ar > 0) * cc.shape + (cor_ar + maxsize_ar <= 0) * (shape_ar - z1end + z1beg) - 1
            z2end = z2beg + z1end - z1beg

            if pmcc:
                std_z1p = z1p.std()
            if normalize == 2:
                z1p /= z1p.mean()

            for k in range(cc.shape[0]):
                for l in range(cc.shape[1]):
                    if pmcc:
                        z2p = z2[z2beg[0] - k : z2end[0] - k, z2beg[1] - l : z2end[1] - l].flatten()
                        std_z2p = z2p.std()
                        cc[k, l] = -np.cov(z1p, z2p, rowvar=True)[1, 0] / (std_z1p * std_z2p)
                    else:
                        if normalize == 2:
                            z2p = z2[z2beg[0] - k : z2end[0] - k, z2beg[1] - l : z2end[1] - l].flatten()
                            z2p /= z2p.mean()
                            z2p -= z1p
                        else:
                            z2p = z2[z2beg[0] - k : z2end[0] - k, z2beg[1] - l : z2end[1] - l].flatten()
                            z2p -= z1p
                        cc[k, l] = ((z2p - z2p.mean()) ** 2).sum()
                        # cc(k,l) = std(z1p./z2(z2beg(1)-k:z2end(1)-k,z2beg(2)-l:z2end(2)-l)(:));

            c = np.unravel_index(np.argmin(cc, axis=None), shape=cc.shape)

            if not np.sum((c == 0) + (c == np.array(cc.shape) - 1)):
                # check that we are not at the edge of the region that was sampled
                x = np.array([-1, 0, 1])
                tmp = self.refine_max_position_2d(cc[c[0] - 1 : c[0] + 2, c[1] - 1 : c[1] + 2], x, x)
                c += tmp
            shiftfound = True

            c += z1beg - z2beg

            rapp_hist = []
            if not shiftfound:
                cor_estimate = c
                # Check that new shift estimate was not already done (avoid eternal loop)
                if self._checkifpart(cor_estimate, rapp_hist):
                    self.logger.debug("Stuck in loop?")
                    refine = True
                    shiftfound = True
                    c = np.array([np.nan, np.nan])
                else:
                    rapp_hist.append(cor_estimate)
                    self.logger.debug(f"Changing shift estimate: {cor_estimate}")
                    maxsize = np.minimum(maxsize, np.array(z1.shape) - np.abs(cor_estimate) - 1).astype(int)
                    if (maxsize == 0).sum():
                        self.logger.debug("Edge of image reached")
                        refine = False
                        shiftfound = True
                        c = np.array([np.nan, np.nan])
            elif len(rapp_hist) > 0:
                self.logger.debug("\n")

        ####################################
        # refine result; useful when shifts are not integer values
        # JL: I don't understand why this refine step should be useful.
        # In Octave, from fastomo.m, refine is always set to False.
        # So this could be ignored.
        # I keep it for future use if it proves useful.
        # if refine:
        #    if debug:
        #        print('Refining solution ...')
        #    z2n = self.interpolate(z2,c)
        #    indices = np.ceil(np.abs(c)).astype(int)
        #    z1p = np.roll(z1,((c>0) * (-1) * indices),[0,1])
        #    z1p = z1p[1:-indices[0]-1,1:-indices[1]-1].flatten()
        #    z2n = np.roll(z2n,((c>0) * (-1) * indices),[0,1])
        #    z2n = z2n[:-indices[0],:-indices[1]]
        #    ccrefine = np.zeros([3,3])
        #    [n2,m2] = z2n.shape
        #    for k in range(3):
        #        for l in range(3):
        #            z2p = z1p - z2n[2-k:n2-k,2-l:m2-l].flatten()
        #            ccrefine[k,l] = ((z2p - z2p.mean())**2).sum()
        #    x = np.array([-1,0,1])
        #    crefine = self.refine_max_position_2d(ccrefine, x, x)
        #    #crefine = min2par(ccrefine)

        #    # Check if the refinement is effectively confined to subpixel
        #    if (np.abs(crefine) >= 1).sum():
        #        self.logger.info("Problems refining result\n")
        #    else:
        #        c += crefine

        return c

    def find_shift(
        self,
        img_1,
        img_2,
        side="center",
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        low_pass=0.01,
        high_pass=None,
        maxsize=[5, 5],
        refine=None,
        pmcc=False,
        normalize=True,
        limz=0.5,
        return_relative_to_middle=None,
    ):
        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotationOctaveAccurate.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---
        self._check_img_pair_sizes(img_1, img_2)

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        cc = self._compute_correlation_fft(
            img_1,
            img_2,
            padding_mode,
            high_pass=high_pass,
            low_pass=low_pass,
        )

        # We use fftshift to deal more easily with negative shifts.
        # This has a cost of subtracting half the image shape afterward.
        shift = np.unravel_index(np.argmax(np.fft.fftshift(cc)), img_shape)
        shift -= np.array(img_shape) // 2

        # The real "accurate" starts here (i.e. the octave findshift() func).
        if np.abs(shift[0]) > 10 * limz:
            # This is suspiscious. We don't trust results of correlate.
            self.logger.warning("Pre-correlation yields {shift[0]} pixels vertical motion")
            self.logger.warning("We do not consider it.")
            shift = (0, 0)

        # Limit the size of region for comparison to cutsize in both directions.
        # Hard-coded?
        cutsize = img_shape[1] // 2
        oldshift = np.round(shift).astype(int)
        if (img_shape[0] > cutsize) or (img_shape[1] > cutsize):
            im0 = self._cut(img_1, min(img_shape[0], cutsize), min(img_shape[1], cutsize))
            im1 = self._cut(
                np.roll(img_2, oldshift, axis=(0, 1)), min(img_shape[0], cutsize), min(img_shape[1], cutsize)
            )
            shift = oldshift + self._local_correlation(
                im0,
                im1,
                maxsize=maxsize,
                refine=refine,
                pmcc=pmcc,
                normalize=normalize,
            )
        else:
            shift = self._local_correlation(
                img_1,
                img_2,
                maxsize=maxsize,
                cor_estimate=oldshift,
                refine=refine,
                pmcc=pmcc,
                normalize=normalize,
            )
        if ((shift - oldshift) ** 2).sum() > 4:
            self.logger.warning(f"Pre-correlation ({oldshift}) and accurate correlation ({shift}) are not consistent.")
            self.logger.warning("Please check!!!")

        offset = shift[1] / 2

        if np.abs(shift[0]) > limz:
            self.logger.debug("Verify alignment or sample motion.")
            self.logger.debug(f"Verical motion: {shift[0]} pixels.")
            self.logger.debug(f"Offset?: {offset} pixels.")
        else:
            self.logger.debug(f"Offset?: {offset} pixels.")

        if not (return_relative_to_middle):
            offset += (img_shape[1] - 1) / 2

        return offset


# COMPAT.
from .cor_sino import CenterOfRotationFourierAngles as CenterOfRotationFourierAngles0

CenterOfRotationFourierAngles = deprecated_class(
    "CenterOfRotationFourierAngles was moved to nabu.estimation.cor_sino", do_print=True
)(CenterOfRotationFourierAngles0)
#



import os
from typing import Optional
import contextlib
import h5py
import numpy as np
from silx.io.url import DataUrl
from tomoscan.volumebase import VolumeBase
from tomoscan.esrf import EDFVolume, HDF5Volume, TIFFVolume, JP2KVolume, MultiTIFFVolume
from tomoscan.io import HDF5File


# This function might be moved elsewhere
def get_compacted_dataslices(urls, subsampling=None, begin=0):
    """
    Regroup urls to get the data more efficiently.
    Build a structure mapping files indices to information on
    how to load the data: `{indices_set: data_location}`
    where `data_location` contains contiguous indices.

    Parameters
    -----------
    urls: dict
        Dictionary where the key is an integer and the value is a silx `DataUrl`.
    subsampling: int, optional
        Subsampling factor when reading the frames. If an integer `n` is provided,
        then one frame out of `n` will be read.

    Returns
    --------
    merged_urls: dict
        Dictionary with the same keys as the `urls` parameter, and where the
        values are the corresponding `silx.io.url.DataUrl` with merged data_slice.
    """
    subsampling = subsampling or 1

    def _convert_to_slice(idx):
        if np.isscalar(idx):
            return slice(idx, idx + 1)
        # otherwise, assume already slice object
        return idx

    def is_contiguous_slice(slice1, slice2, step=1):
        if np.isscalar(slice1):
            slice1 = slice(slice1, slice1 + step)
        if np.isscalar(slice2):
            slice2 = slice(slice2, slice2 + step)
        return slice2.start == slice1.stop

    def merge_slices(slice1, slice2, step=1):
        return slice(slice1.start, slice2.stop, step)

    if len(urls) == 0:
        return urls

    sorted_files_indices = sorted(urls.keys())
    # if begin > 0:
    # sorted_files_indices = sorted_files_indices[begin:]
    idx0 = sorted_files_indices[begin]
    first_url = urls[idx0]

    merged_indices = [[idx0]]
    # location = (file_path, data_path, slice)
    data_location = [[first_url.file_path(), first_url.data_path(), _convert_to_slice(first_url.data_slice())]]
    pos = 0
    curr_fp, curr_dp, curr_slice = data_location[pos]
    skip_next = 0
    for idx in sorted_files_indices[begin + 1 :]:
        if skip_next > 1:
            skip_next -= 1
            continue
        url = urls[idx]
        next_slice = _convert_to_slice(url.data_slice())
        if (
            (url.file_path() == curr_fp)
            and (url.data_path() == curr_dp)
            and is_contiguous_slice(curr_slice, next_slice, step=subsampling)
        ):
            merged_indices[pos].append(idx)
            merged_slices = merge_slices(curr_slice, next_slice, step=subsampling)
            data_location[pos][-1] = merged_slices
            curr_slice = merged_slices
            skip_next = 0
        else:  # "jump"
            if begin > 0 and skip_next == 0:
                # Skip the "begin" next urls (first of a new block)
                skip_next = begin
                continue
            pos += 1
            merged_indices.append([idx])
            data_location.append([url.file_path(), url.data_path(), _convert_to_slice(url.data_slice())])
            curr_fp, curr_dp, curr_slice = data_location[pos]

    # Format result
    res = {}
    for ind, dl in zip(merged_indices, data_location):
        res.update(dict.fromkeys(ind, DataUrl(file_path=dl[0], data_path=dl[1], data_slice=dl[2])))

    return res


def get_first_hdf5_entry(fname):
    with HDF5File(fname, "r") as fid:
        entry = list(fid.keys())[0]
    return entry


def hdf5_entry_exists(fname, entry):
    with HDF5File(fname, "r") as fid:
        res = fid.get(entry, None) is not None
    return res


def get_h5_value(fname, h5_path, default_ret=None):
    with HDF5File(fname, "r") as fid:
        try:
            val_ptr = fid[h5_path][()]
        except KeyError:
            val_ptr = default_ret
    return val_ptr


def get_h5_str_value(dataset_ptr):
    """
    Get a HDF5 field which can be bytes or str (depending on h5py version !).
    """
    data = dataset_ptr[()]
    if isinstance(data, str):
        return data
    else:
        return bytes.decode(data)


def create_dict_of_indices(images_stack, images_indices):
    """
    From an image stack with the images indices, create a dictionary where
    each index is the image index, and the value is the corresponding image.

    Parameters
    ----------
    images_stack: numpy.ndarray
        A 3D numpy array in the layout (n_images, n_y, n_x)
    images_indices: array or list of int
        Array containing the indices of images in the stack

    Examples
    --------
    Given a simple array stack:

    >>> images_stack = np.arange(3*4*5).reshape((3,4,5))
    ... images_indices = [2, 7, 1]
    ... create_dict_of_indices(images_stack, images_indices)
    ... # returns {2: array1, 7: array2, 1: array3}
    """
    if images_stack.ndim != 3:
        raise ValueError("Expected a 3D array")
    if len(images_indices) != images_stack.shape[0]:
        raise ValueError("images_stack must have as many images as the length of images_indices")
    res = {}
    for i in range(len(images_indices)):
        res[images_indices[i]] = images_stack[i]
    return res


def convert_dict_values(dic, val_replacements, bytes_tostring=False):
    """
    Modify a dictionary to be able to export it with silx.io.dicttoh5
    """
    modified_dic = {}
    for key, value in dic.items():
        if isinstance(key, int):  # np.isscalar ?
            key = str(key)
        if isinstance(value, bytes) and bytes_tostring:
            value = bytes.decode(value.tostring())
        if isinstance(value, dict):
            value = convert_dict_values(value, val_replacements, bytes_tostring=bytes_tostring)
        else:
            if isinstance(value, DataUrl):
                value = value.path()
            elif value.__hash__ is not None and value in val_replacements:
                value = val_replacements[value]
        modified_dic[key] = value
    return modified_dic


class _BaseReader(contextlib.AbstractContextManager):
    def __init__(self, url: DataUrl):
        if not isinstance(url, DataUrl):
            raise TypeError("url should be an instance of DataUrl")
        if url.scheme() not in ("silx", "h5py"):
            raise ValueError("Valid scheme are silx and h5py")
        if url.data_slice() is not None:
            raise ValueError("Data slices are not managed. Data path should " "point to a bliss node (h5py.Group)")
        self._url = url
        self._file_handler = None

    def __exit__(self, *exc):
        return self._file_handler.close()


class EntryReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = HDF5File(self._url.file_path(), mode="r")
        if self._url.data_path() == "":
            entry = self._file_handler
        else:
            entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Group):
            raise ValueError("Data path should point to a bliss node (h5py.Group)")
        return entry


class DatasetReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = HDF5File(self._url.file_path(), mode="r")
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Dataset):
            raise ValueError("Data path ({}) should point to a dataset (h5py.Dataset)".format(self._url.path()))
        return entry


# TODO: require some utils function to deduce type. And insure homogeneity. Might be moved in tomoscan ?
def file_format_is_edf(file_format: str):
    return file_format.lower().lstrip(".") == "edf"


def file_format_is_jp2k(file_format: str):
    return file_format.lower().lstrip(".") in ("jp2k", "jp2")


def file_format_is_tiff(file_format: str):
    return file_format.lower().lstrip(".") in ("tiff", "tif")


def file_format_is_hdf5(file_format: str):
    return file_format.lower().lstrip(".") in ("hdf5", "hdf", "nx", "nexus")


def get_output_volume(location: str, file_prefix: Optional[str], file_format: str, multitiff=False) -> VolumeBase:
    # TODO: see strategy. what if user provide a .nx ... ?
    # this function should be more generic
    location, extension = os.path.splitext(location)
    if extension == "":
        extension = file_format
    if file_format_is_edf(extension):
        return EDFVolume(folder=location, volume_basename=file_prefix)
    elif file_format_is_jp2k(extension):
        return JP2KVolume(folder=location, volume_basename=file_prefix)
    elif file_format_is_hdf5(file_format=extension):
        if extension is None:
            if file_prefix is None:
                location = ".".join([location, extension])
            else:
                location = os.path.join(location, ".".join([file_prefix, extension]))
        return HDF5Volume(file_path=location)
    elif file_format_is_tiff(extension):
        if multitiff:
            return MultiTIFFVolume(file_path=location)
        else:
            return TIFFVolume(folder=location, volume_basename=file_prefix)



import numpy as np
from ..utils import get_opencl_srcfile, get_cuda_srcfile, updiv, BaseClassError, MissingComponentError
from ..opencl.utils import __has_pyopencl__
from ..cuda.utils import __has_pycuda__

if __has_pyopencl__:
    from ..opencl.kernel import OpenCLKernel
    from ..opencl.processing import OpenCLProcessing
    from pyopencl.tools import dtype_to_ctype as cl_dtype_to_ctype
else:
    OpenCLKernel = OpenCLProcessing = cl_dtype_to_ctype = MissingComponentError("need pyopencl to use this class")
if __has_pycuda__:
    from ..cuda.kernel import CudaKernel
    from ..cuda.processing import CudaProcessing
    from pycuda.tools import base_dtype_to_ctype as cu_dtype_to_ctype
else:
    CudaKernel = CudaProcessing = cu_dtype_to_ctype = MissingComponentError("need pycuda to use this class")


# pylint: disable=E1101, E1102
class TransposeBase:
    """
    A class for transposing (out-of-place) a cuda or opencl array
    """

    KernelCls = BaseClassError
    ProcessingCls = BaseClassError
    dtype_to_ctype = BaseClassError
    backend = "none"

    def __init__(self, shape, dtype, dst_dtype=None, **backend_options):
        self.processing = self.ProcessingCls(**(backend_options or {}))
        self.shape = shape
        self.dtype = dtype
        self.dst_dtype = dst_dtype or dtype
        if len(shape) != 2:
            raise ValueError("Expected 2D array")

        self._kernel_init_args = [
            "transpose",
        ]
        self._kernel_init_kwargs = {
            "options": [
                "-DSRC_DTYPE=%s" % self.dtype_to_ctype(self.dtype),
                "-DDST_DTYPE=%s" % self.dtype_to_ctype(self.dst_dtype),
            ],
        }
        self._configure_kenel_initialization()
        self._transpose_kernel = self.KernelCls(*self._kernel_init_args, **self._kernel_init_kwargs)
        self._configure_kernel_call()

    def __call__(self, arr, dst=None):
        if dst is None:
            dst = self.processing.allocate_array("dst", self.shape[::-1], dtype=self.dst_dtype)
        self._transpose_kernel(arr, dst, np.int32(self.shape[1]), np.int32(self.shape[0]), **self._kernel_kwargs)
        return dst


class CudaTranspose(TransposeBase):
    KernelCls = CudaKernel
    ProcessingCls = CudaProcessing
    dtype_to_ctype = cu_dtype_to_ctype
    backend = "cuda"

    def _configure_kenel_initialization(self):
        self._kernel_init_kwargs.update(
            {
                "filename": get_cuda_srcfile("transpose.cu"),
                "signature": "PPii",
            }
        )

    def _configure_kernel_call(self):
        block = (32, 32, 1)
        grid = [updiv(a, b) for a, b in zip(self.shape, block)]
        self._kernel_kwargs = {"grid": grid, "block": block}


class OpenCLTranspose(TransposeBase):
    KernelCls = OpenCLKernel
    ProcessingCls = OpenCLProcessing
    dtype_to_ctype = cl_dtype_to_ctype
    backend = "opencl"

    def _configure_kenel_initialization(self):
        self._kernel_init_args.append(self.processing.ctx)
        self._kernel_init_kwargs.update(
            {
                "filename": get_opencl_srcfile("transpose.cl"),
                "queue": self.processing.queue,
            }
        )

    def _configure_kernel_call(self):
        block = (16, 16, 1)
        grid = [updiv(a, b) * b for a, b in zip(self.shape, block)]
        self._kernel_kwargs = {"global_size": grid, "local_size": block}


#
# An attempt to have a simplified access to transpose operation
#

# (backend, shape, dtype, dtype_out)
_transposes_store = {}


def transpose(array, dst=None, **backend_options):
    if hasattr(array, "with_queue"):
        backend = "opencl"
        transpose_cls = OpenCLTranspose
        backend_options["queue"] = array.queue  # !
    elif hasattr(array, "bind_to_texref"):
        backend = "cuda"
        transpose_cls = CudaTranspose
    else:
        raise ValueError("array should be either a pycuda.gpuarray.GPUArray or pyopencl.array.Array instance")

    dst_dtype = dst.dtype if dst is not None else None
    key = (backend, array.shape, np.dtype(array.dtype), dst_dtype)
    transpose_instance = _transposes_store.get(key, None)
    if transpose_instance is None:
        transpose_instance = transpose_cls(array.shape, array.dtype, dst_dtype=dst_dtype, **backend_options)
        _transposes_store[key] = transpose_instance

    return transpose_instance(array, dst=dst)



from ..processing.padding_opencl import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.opencl.padding has been moved to nabu.processing.padding_opencl", do_print=True, func_name="padding_opencl"
)



import pytest
import numpy as np
import scipy.interpolate
from nabu.processing.fft_cuda import get_available_fft_implems
from nabu.testutils import get_data as nabu_get_data
from nabu.testutils import __do_long_tests__
from nabu.preproc.flatfield import FlatFieldArrays
from nabu.preproc.ccd import CCDFilter
from nabu.preproc import ctf
from nabu.estimation.distortion import estimate_flat_distortion
from nabu.misc.filters import correct_spikes
from nabu.preproc.distortion import DistortionCorrection
from nabu.cuda.utils import __has_pycuda__, get_cuda_context

__has_cufft__ = False
if __has_pycuda__:
    from nabu.preproc.ctf_cuda import CudaCTFPhaseRetrieval

    avail_fft = get_available_fft_implems()
    __has_cufft__ = len(avail_fft) > 0


@pytest.fixture(scope="class")
def bootstrap_TestCtf(request):
    cls = request.cls
    cls.abs_tol = 1.0e-4

    test_data = nabu_get_data("ctf_tests_data_all_pars.npz")

    cls.rand_disp_vh = test_data["rh"]
    ## the dimension number 1 is over holotomo distances, so far our filter is for one distance only
    cls.rand_disp_vh.shape = [cls.rand_disp_vh.shape[0], cls.rand_disp_vh.shape[2]]
    cls.dark = test_data["dark"]
    cls.flats = [test_data["ref0"], test_data["ref1"]]
    cls.im = test_data["im"]
    cls.ipro = int(test_data["ipro"])
    cls.expected_result = test_data["result"]
    cls.ref_plain = test_data["ref_plain_float_flat"]

    cls.flats_n = test_data["refns"]

    cls.img_shape_vh = test_data["img_shape_vh"]
    cls.padded_img_shape_vh = test_data["padded_img_shape_vh"]
    cls.z1_vh = test_data["z1_vh"]
    cls.z2 = test_data["z2"]
    cls.pix_size_det = test_data["pix_size_det"][()]
    cls.length_scale = test_data["length_scale"]
    cls.wavelength = test_data["wave_length"]
    cls.remove_spikes_threshold = test_data["remove_spikes_threshold"]
    cls.delta_beta = 27


@pytest.mark.usefixtures("bootstrap_TestCtf")
class TestCtf:
    def check_result(self, res, ref, error_message):
        diff = np.abs(res - ref)
        diff[diff > np.percentile(diff, 99)] = 0
        assert diff.max() < self.abs_tol * (np.abs(ref).mean()), error_message

    def test_ctf_id16_way(self):
        """test the ctf phase retrieval.
        The cft filter, of the CtfFilter class is iniitalised with the geomety informations contained in geo_pars object
        of the GeoPars class.
        The geometry encompass the case of astigmatic wavefront with a vertical and horisontal sources which are
        at distance z1_vh[0] and z1_vh[1] from the object.
        In the case of parllel geometry put  z1_vh[0]= z1_vh[1] = R
        where R is a large value ( meters).
        SI unit system is used. But the same results shudl be obtained with any homogenuous choice of the distance units.
        The img_shape is the shape of the images which will be processed.
        padded_img_shape is an intermediate shape which needs to be larger that the img_shape to avoid border effect due
        to convolutions.
        Length scale is an internal parameters which should not affect in anyway the result unless there are serious numerical
        problems involving very small lenghts. You can safely let the default value.
        """
        geo_pars = ctf.GeoPars(
            z1_vh=self.z1_vh,
            z2=self.z2,
            pix_size_det=self.pix_size_det,
            length_scale=self.length_scale,
            wavelength=self.wavelength,
        )

        flats = FlatFieldArrays(
            [1200] + list(self.img_shape_vh), {0: self.flats[0], 1200: self.flats[1]}, {0: self.dark}
        )

        my_flat = flats.get_flat(self.ipro)
        my_img = self.im - self.dark
        my_flat = my_flat - self.dark

        new_coordinates = estimate_flat_distortion(
            my_flat,
            my_img,
            tile_size=100,
            interpolation_kind="cubic",
            padding_mode="edge",
            correction_spike_threshold=3,
        )

        interpolator = scipy.interpolate.RegularGridInterpolator(
            (np.arange(my_flat.shape[0]), np.arange(my_flat.shape[1])),
            my_flat,
            bounds_error=False,
            method="linear",
            fill_value=None,
        )
        my_flat = interpolator(new_coordinates)

        my_img = my_img / my_flat
        my_img = correct_spikes(my_img, self.remove_spikes_threshold)

        my_shift = self.rand_disp_vh[:, self.ipro]

        ctf_filter = ctf.CtfFilter(
            self.dark.shape,
            geo_pars,
            self.delta_beta,
            padded_shape=self.padded_img_shape_vh,
            translation_vh=my_shift,
            normalize_by_mean=True,
            lim1=1.0e-5,
            lim2=0.2,
        )
        phase = ctf_filter.retrieve_phase(my_img)

        self.check_result(
            phase, self.expected_result, "retrieved phase and reference result differ beyond the accepted tolerance"
        )

    @pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
    def test_ctf_id16_class(self):
        geo_pars = ctf.GeoPars(
            z1_vh=self.z1_vh,
            z2=self.z2,
            pix_size_det=self.pix_size_det,
            length_scale=self.length_scale,
            wavelength=self.wavelength,
        )
        distortion_correction = DistortionCorrection(
            estimation_method="fft-correlation",
            estimation_kwargs={
                "tile_size": 100,
                "interpolation_kind": "cubic",
                "padding_mode": "edge",
                "correction_spike_threshold": 3.0,
            },
            correction_method="interpn",
            correction_kwargs={"fill_value": None},
        )
        flats = FlatFieldArrays(
            [1200] + list(self.img_shape_vh),
            {0: self.flats[0], 1200: self.flats[1]},
            {0: self.dark},
            distortion_correction=distortion_correction,
        )

        # The "correct_spikes" function is numerically unstable (comparison with a float threshold).
        # If float32 is used for the image, one spike is detected while it is not in the previous test
        # (although the max difference between the inputs is about 1e-8).
        # We use float64 data type for the image to make tests pass.
        img = self.im.astype(np.float64)

        flats.normalize_single_radio(img, self.ipro)
        img = correct_spikes(img, self.remove_spikes_threshold)

        shift = self.rand_disp_vh[:, self.ipro]

        ctf_filter = ctf.CtfFilter(
            img.shape,
            geo_pars,
            self.delta_beta,
            padded_shape=self.padded_img_shape_vh,
            translation_vh=shift,
            normalize_by_mean=True,
            lim1=1.0e-5,
            lim2=0.2,
        )
        phase = ctf_filter.retrieve_phase(img)

        message = "retrieved phase and reference result differ beyond the accepted tolerance"
        assert np.abs(phase - self.expected_result).max() < 10 * self.abs_tol * (
            np.abs(self.expected_result).mean()
        ), message

    def test_ctf_plain_way(self):
        geo_pars = ctf.GeoPars(
            z1_vh=None,
            z2=self.z2,
            pix_size_det=self.pix_size_det,
            length_scale=self.length_scale,
            wavelength=self.wavelength,
        )
        flatfielder = FlatFieldArrays(
            [1] + list(self.img_shape_vh),
            {0: self.flats[0], 1200: self.flats[1]},
            {0: self.dark},
            radios_indices=[self.ipro],
        )

        spikes_corrector = CCDFilter(
            self.dark.shape, median_clip_thresh=self.remove_spikes_threshold, abs_diff=True, preserve_borders=True
        )

        img = self.im.astype("f")
        img = flatfielder.normalize_radios(np.array([img]))[0]
        img = spikes_corrector.median_clip_correction(img)

        ctf_args = [img.shape, geo_pars, self.delta_beta]
        ctf_kwargs = {"padded_shape": self.padded_img_shape_vh, "normalize_by_mean": True, "lim1": 1.0e-5, "lim2": 0.2}
        ctf_filter = ctf.CtfFilter(*ctf_args, **ctf_kwargs)
        phase = ctf_filter.retrieve_phase(img)

        self.check_result(phase, self.ref_plain, "Something wrong with CtfFilter")

        # Test R2C
        ctf_numpy = ctf.CtfFilter(*ctf_args, **ctf_kwargs, use_rfft=True)
        phase_r2c = ctf_numpy.retrieve_phase(img)
        self.check_result(phase_r2c, self.ref_plain, "Something wrong with CtfFilter-R2C")

        # Test multi-core FFT
        ctf_fft = ctf.CtfFilter(*ctf_args, **ctf_kwargs, use_rfft=True, fft_num_threads=0)
        if ctf_fft.use_rfft:
            phase_fft = ctf_fft.retrieve_phase(img)
            self.check_result(phase_r2c, self.ref_plain, "Something wrong with CtfFilter-FFT")

    @pytest.mark.skipif(not (__has_pycuda__ and __has_cufft__), reason="pycuda and (scikit-cuda or vkfft)")
    def test_cuda_ctf(self):
        data = nabu_get_data("brain_phantom.npz")["data"]
        delta_beta = 50.0
        energy_kev = 22.0
        distance_m = 1.0
        pix_size_m = 0.1e-6

        geo_pars = ctf.GeoPars(z2=distance_m, pix_size_det=pix_size_m, wavelength=1.23984199e-9 / energy_kev)
        ctx = get_cuda_context()

        for normalize in [True, False]:
            ctf_filter = ctf.CTFPhaseRetrieval(
                data.shape, geo_pars, delta_beta=delta_beta, normalize_by_mean=normalize, use_rfft=True
            )
            cuda_ctf_filter = CudaCTFPhaseRetrieval(
                data.shape,
                geo_pars,
                delta_beta=delta_beta,
                use_rfft=True,
                normalize_by_mean=normalize,
            )
            ref = ctf_filter.retrieve_phase(data)

            d_data = cuda_ctf_filter.cuda_processing.to_device("_d_data", data)
            res = cuda_ctf_filter.retrieve_phase(d_data).get()
            err_max = np.max(np.abs(res - ref))

            assert err_max < 1e-2, "Something wrong with retrieve_phase(normalize_by_mean=%s)" % (str(normalize))



import pytest
from nabu.stitching.frame_composition import FrameComposition
import numpy

from nabu.stitching.overlap import OverlapStitchingStrategy, ImageStichOverlapKernel


def test_frame_composition():
    """
    Test FrameComposition
    """
    frame_0 = numpy.zeros((100, 1))
    frame_1 = numpy.ones((98, 1))
    frame_2 = numpy.ones((205, 1)) * 2.0

    frames = (frame_0, frame_1, frame_2)

    y_shifts = -20, -10

    kernels = [
        ImageStichOverlapKernel(
            stitching_axis=0,
            frame_unstitched_axis_size=1,
            stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
            overlap_size=4,
        ),
        ImageStichOverlapKernel(
            stitching_axis=0,
            frame_unstitched_axis_size=1,
            stitching_strategy=OverlapStitchingStrategy.MEAN,
            overlap_size=8,
        ),
    ]
    # check raw composition
    raw_composition = FrameComposition.compute_raw_frame_compositions(
        frames=frames,
        key_lines=(
            (90, 10),
            (98 - 5, 5),
        ),
        overlap_kernels=kernels,
        stitching_axis=0,
    )
    assert isinstance(raw_composition, FrameComposition)

    assert raw_composition.local_start == (0, 12, 9)
    assert raw_composition.local_end == (88, 89, 205)
    assert raw_composition.global_start == (0, 92, 177)
    assert raw_composition.global_end == (88, 169, 373)

    stitched_data = numpy.empty((100 + 98 + 205 - 30, 1))
    raw_composition.compose(output_frame=stitched_data, input_frames=frames)
    assert stitched_data[0, 0] == 0
    assert stitched_data[150, 0] == 1.0
    assert stitched_data[-1, 0] == 2.0

    # check stitch composition
    stitch_composition = FrameComposition.compute_stitch_frame_composition(
        frames=frames,
        key_lines=(
            (90, 10),
            (98 - 5, 5),
        ),
        overlap_kernels=kernels,
        stitching_axis=0,
    )

    FrameComposition.pprint_composition(raw_composition, stitch_composition)

    assert stitch_composition.local_start == (0, 0)
    assert stitch_composition.local_end == (4, 8)
    assert stitch_composition.global_start == (88, 169)
    assert stitch_composition.global_end == (92, 177)

    stitched_frames = []
    for frame_0, frame_1, kernel, y_shift in zip(frames[:-1], frames[1:], kernels, y_shifts):
        # take frames once shifted
        frame_0_overlap = frame_0[y_shift:]
        frame_1_overlap = frame_1[:-y_shift]
        # select the overlap area
        frame_0_overlap = frame_0[-kernel.overlap_size :]
        frame_1_overlap = frame_1[: kernel.overlap_size]

        stitched_frames.append(kernel.stitch(frame_0_overlap, frame_1_overlap)[0])

    stitch_composition.compose(
        output_frame=stitched_data,
        input_frames=stitched_frames,
    )
    assert 0.0 < stitched_data[90, 0] < 1.0
    assert 1.0 < stitched_data[172, 0] < 2.0


_raw_comp_config = (
    {
        "key_lines": (
            (17, 2),
            (36, 3),
        ),
        "raw_global_start": (0, 19, 53),
        "raw_global_end": (16, 49, 68),
        "raw_local_start": (0, 4, 5),
        "raw_local_end": (16, 34, 20),
        "kernels": (
            ImageStichOverlapKernel(
                stitching_axis=0,
                frame_unstitched_axis_size=1,
                stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
                overlap_size=3,
            ),
            ImageStichOverlapKernel(
                stitching_axis=0,
                frame_unstitched_axis_size=1,
                stitching_strategy=OverlapStitchingStrategy.MEAN,
                overlap_size=4,
            ),
        ),
    },
)


@pytest.mark.parametrize("configuration", _raw_comp_config)
def test_raw_frame_composition_exotic_config(configuration):
    """
    Test some
    """
    frame_0 = numpy.zeros((20, 1))
    frame_1 = numpy.ones((40, 1))
    frame_2 = numpy.ones((20, 1)) * 2.0

    frames = (frame_0, frame_1, frame_2)

    key_lines = configuration.get("key_lines")
    kernels = configuration.get("kernels")
    # check raw composition
    raw_composition = FrameComposition.compute_raw_frame_compositions(
        frames=frames,
        overlap_kernels=kernels,
        key_lines=key_lines,
        stitching_axis=0,
    )

    assert raw_composition.global_start == configuration.get("raw_global_start")
    assert raw_composition.global_end == configuration.get("raw_global_end")
    assert raw_composition.local_start == configuration.get("raw_local_start")
    assert raw_composition.local_end == configuration.get("raw_local_end")

    stitched_data = numpy.empty(
        (
            (raw_composition.global_end[-1] - raw_composition.global_start[0]),
            1,
        )
    )
    raw_composition.compose(output_frame=stitched_data, input_frames=frames)



import os
import numpy as np
from nxtomo.io import HDF5File
from silx.io.url import DataUrl
from silx.io import get_data
from tomoscan.framereducer.reducedframesinfos import ReducedFramesInfos
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from ..utils import check_supported, is_writeable


def get_frame_possible_urls(dataset_info, user_dir, output_dir):
    """
    Return a dict with the possible location of reduced dark/flat frames.

    Parameters
    ----------
    dataset_info: DatasetAnalyzer object
        DatasetAnalyzer object: data structure containing information on the parsed dataset
    user_dir: str or None
        User-provided directory location for the reduced frames.
    output_dir: str or None
        Output processing directory
    """

    frame_types = ["flats", "darks"]
    h5scan = dataset_info.dataset_scanner  # tomoscan object

    def make_dataurl(dirname, frame_type):
        """
        The template formatting should be done by tomoscan in principle, but this complicates logging.
        """

        if frame_type == "flats":
            dataurl_default_template = h5scan.REDUCED_FLATS_DATAURLS[0]
        else:
            dataurl_default_template = h5scan.REDUCED_DARKS_DATAURLS[0]

        rel_file_path = dataurl_default_template.file_path().format(scan_prefix=h5scan.get_dataset_basename())
        return DataUrl(
            file_path=os.path.join(dirname, rel_file_path),
            data_path=dataurl_default_template.data_path().format(entry=h5scan.entry, index="{index}"),
            data_slice=dataurl_default_template.data_slice(),  # not sure if needed
            scheme="silx",
        )

    urls = {"user": None, "dataset": None, "output": None}

    if user_dir is not None:
        urls["user"] = {frame_type: make_dataurl(user_dir, frame_type) for frame_type in frame_types}

    # tomoscan.esrf.scan.hdf5scan.REDUCED_{DARKS|FLATS}_DATAURLS.file_path() is a relative path
    # Create a absolute path instead
    urls["dataset"] = {
        frame_type: make_dataurl(os.path.dirname(h5scan.master_file), frame_type) for frame_type in frame_types
    }

    if output_dir is not None:
        urls["output"] = {frame_type: make_dataurl(output_dir, frame_type) for frame_type in frame_types}

    return urls


def save_reduced_frames(dataset_info, reduced_frames_arrays, reduced_frames_urls):
    reduce_func = {"flats": np.median, "darks": np.mean}  # TODO configurable ?

    # Get "where to write". tomoscan expects a DataUrl
    darks_flats_dir_url = reduced_frames_urls.get("user", None)
    if darks_flats_dir_url is not None:
        output_url = darks_flats_dir_url
    elif is_writeable(os.path.dirname(reduced_frames_urls["dataset"]["flats"].file_path())):
        output_url = reduced_frames_urls["dataset"]
    else:
        output_url = reduced_frames_urls["output"]

    # Get the "ReducedFrameInfos" data structure expected by tomoscan
    def _get_additional_info(frame_type):
        electric_current = dataset_info.dataset_scanner.electric_current
        count_time = dataset_info.dataset_scanner.count_time
        if electric_current is not None:
            electric_current = {
                sl.start: reduce_func[frame_type](electric_current[sl]) for sl in dataset_info.frames_slices(frame_type)
            }
            electric_current = [electric_current[k] for k in sorted(electric_current.keys())]
        if count_time is not None:
            count_time = {
                sl.start: reduce_func[frame_type](count_time[sl]) for sl in dataset_info.frames_slices(frame_type)
            }
            count_time = [count_time[k] for k in sorted(count_time.keys())]
        info = ReducedFramesInfos()
        info.count_time = count_time
        info.machine_electric_current = electric_current
        return info

    flats_info = _get_additional_info("flats")
    darks_info = _get_additional_info("darks")

    # Call tomoscan to save the reduced frames
    dataset_info.dataset_scanner.save_reduced_darks(
        reduced_frames_arrays["darks"],
        output_urls=[output_url["darks"]],
        darks_infos=darks_info,
        metadata_output_urls=[get_metadata_url(output_url["darks"], "darks")],
        overwrite=True,
    )
    dataset_info.dataset_scanner.save_reduced_flats(
        reduced_frames_arrays["flats"],
        output_urls=[output_url["flats"]],
        flats_infos=flats_info,
        metadata_output_urls=[get_metadata_url(output_url["flats"], "flats")],
        overwrite=True,
    )
    dataset_info.logger.info("Saved reduced darks/flats to %s" % output_url["flats"].file_path())
    return output_url, flats_info, darks_info


def get_metadata_url(url, frame_type):
    """
    Return the url of the metadata stored alongside flats/darks
    """
    check_supported(frame_type, ["flats", "darks"], "frame type")
    template_url = getattr(NXtomoScan, "REDUCED_%s_METADATAURLS" % frame_type.upper())[0]
    return DataUrl(
        file_path=url.file_path(),
        data_path=template_url.data_path(),
        scheme="silx",
    )


def tomoscan_load_reduced_frames(dataset_info, frame_type, url):
    tomoscan_method = getattr(dataset_info.dataset_scanner, "load_reduced_%s" % frame_type)
    return tomoscan_method(
        inputs_urls=[url],
        return_as_url=True,
        return_info=True,
        metadata_input_urls=[get_metadata_url(url, frame_type)],
    )


def data_url_exists(data_url):
    """
    Return true iff the file exists and the data URL is valid (i.e data/group is actually in the file)
    """
    if not (os.path.isfile(data_url.file_path())):
        return False
    group_exists = False
    with HDF5File(data_url.file_path(), "r") as f:
        data_path_without_index = data_url.data_path().split("{")[0]
        group_exists = f.get(data_path_without_index, default=None) is not None
    return group_exists


# pylint: disable=E1136
def update_dataset_info_flats_darks(dataset_info, flatfield_mode, output_dir=None, darks_flats_dir=None):
    """
    Update a DatasetAnalyzer object with reduced flats/darks (hereafter "reduced frames").

    How the reduced frames are loaded/computed/saved will depend on the "flatfield_mode" parameter.

    The principle is the following:
    (1) Attempt at loading already-computed reduced frames (XXX_darks.h5 and XXX_flats.h5):
       - First check files in the user-defined directory 'darks_flats_dir'
       - Then try to load from files located alongside the .nx dataset (dataset directory)
       - Then try to load from output_dir, if provided
    (2) If loading fails, or flatfield_mode == "force_compute", compute the reduced frames.
    (3) Save these reduced frames
       - Save in darks_flats_dir, if provided by user
       - Otherwise, save in the data directory (next to the .nx file), if write access OK
       - Otherwise, save in output directory
    """
    if flatfield_mode is False:
        return

    frames_types = ["darks", "flats"]
    reduced_frames_urls = get_frame_possible_urls(dataset_info, darks_flats_dir, output_dir)

    def _compute_and_save_reduced_frames():
        try:
            dataset_info.flats = dataset_info.get_reduced_flats()
            dataset_info.darks = dataset_info.get_reduced_darks()
        except FileNotFoundError:
            msg = "Could not find any flats and/or darks"
            raise FileNotFoundError(msg)
        _, flats_info, darks_info = save_reduced_frames(
            dataset_info, {"darks": dataset_info.darks, "flats": dataset_info.flats}, reduced_frames_urls
        )
        dataset_info.flats_srcurrent = flats_info.machine_electric_current

    if flatfield_mode == "force-compute":
        _compute_and_save_reduced_frames()
        return

    def _can_load_from(folder_type):
        if reduced_frames_urls.get(folder_type, None) is None:
            return False
        return all([data_url_exists(reduced_frames_urls[folder_type][frame_type]) for frame_type in frames_types])

    where_to_load_from = None
    if reduced_frames_urls["user"] is not None and _can_load_from("user"):
        where_to_load_from = "user"
    elif _can_load_from("dataset"):
        where_to_load_from = "dataset"
    elif _can_load_from("output"):
        where_to_load_from = "output"

    if where_to_load_from == None and flatfield_mode == "force-load":
        raise ValueError("Could not load darks/flats (using 'force-load')")

    if where_to_load_from is not None:
        reduced_frames_with_info = {}
        for frame_type in frames_types:
            reduced_frames_with_info[frame_type] = tomoscan_load_reduced_frames(
                dataset_info, frame_type, reduced_frames_urls[where_to_load_from][frame_type]
            )
            dataset_info.logger.info(
                "Loaded %s from %s" % (frame_type, reduced_frames_urls[where_to_load_from][frame_type].file_path())
            )
            red_frames_dict, red_frames_info = reduced_frames_with_info[frame_type]
            setattr(
                dataset_info,
                frame_type,
                {k: get_data(red_frames_dict[k]) for k in red_frames_dict.keys()},
            )
            if frame_type == "flats":
                dataset_info.flats_srcurrent = red_frames_info.machine_electric_current
    else:
        _compute_and_save_reduced_frames()



import math
import numpy as np
from scipy.fft import rfft2, irfft2, fft2, ifft2
from ..resources.logger import LoggerOrPrint
from ..misc import fourier_filters
from ..misc.padding import pad_interpolate, recut
from ..utils import get_num_threads, deprecation_warning


class GeoPars:
    """
    A class to describe the geometry of a phase contrast radiography with
    a source obtained by a focussing system, possibly astigmatic, which is
    at distance z1_vh from the sample. The detector is at z2 from the sample
    """

    def __init__(
        self,
        z1_vh=None,
        z2=None,
        pix_size_det=1e-6,
        wavelength=None,
        magnification=True,
        length_scale=10.0e-6,
        logger=None,
    ):
        """
        Parameters
        ----------
        z1_vh : None, a float, or a sequence of two floats
           the source sample distance (meters), if None the parallel beam is assumed.
           If two floats are given then they are taken as the distance of the vertically focused source (horizontal line)
           and the horizontaly focused source (vertical line) for KB mirrors.
        z2 : float
           the sample detector distance (meters).
        pix_size_det: float or tuple
           pixel size in meters.
           If a tuple is passed, it is interpreted as (horizontal_size, vertical_size)
        wavelength: float
           beam wave length (meters).
        magnification: boolean defaults to True
           if false no magnification is considered
        length_scale: float
            rescaling length scale, meant to avoid having too big or too small numbers.
            defaults to 10.0e-6
        logger: Logger, optional
            A logging object
        """
        self.logger = LoggerOrPrint(logger)
        if z1_vh is None:
            self.z1_vh = None
        else:
            if hasattr(type(z1_vh), "__iter__"):
                self.z1_vh = np.array(z1_vh)
            else:
                self.z1_vh = np.array([z1_vh, z1_vh])
        self.z2 = z2
        self.magnification = magnification
        if np.isscalar(pix_size_det):
            self.pix_size_det_xy = (pix_size_det, pix_size_det)
        else:
            self.pix_size_det_xy = pix_size_det
        self.pix_size_det = self.pix_size_det_xy[0]  # COMPAT

        if self.magnification and self.z1_vh is not None:
            self.M_vh = (self.z1_vh + self.z2) / self.z1_vh
        else:
            self.M_vh = np.array([1, 1])

        self.logger.debug("Magnification : h ({}) ;  v ({}) ".format(self.M_vh[1], self.M_vh[0]))

        self.length_scale = length_scale
        self.wavelength = wavelength

        self.maxM = self.M_vh.max()

        # we bring everything to highest magnification
        self.pix_size_rec_xy = [p / self.maxM for p in self.pix_size_det_xy]
        self.pix_size_rec = self.pix_size_rec_xy[0]  # COMPAT

        which_unit = int(np.sum(np.array([self.pix_size_rec > small for small in [1.0e-6, 1.0e-7]]).astype(np.int32)))
        self.pixelsize_string = [
            "{:.1f} nm".format(self.pix_size_rec * 1e9),
            "{:.3f} um".format(self.pix_size_rec * 1e6),
            "{:.1f} um".format(self.pix_size_rec * 1e6),
        ][which_unit]

        if self.magnification:
            self.logger.debug(
                "All images are resampled to smallest pixelsize: {}".format(self.pixelsize_string),
            )
        else:
            self.logger.debug("Pixelsize images:  {}".format(self.pixelsize_string))


class CTFPhaseRetrieval:
    """
    This class implements the CTF formula of [1] in its regularised form which avoids the zeros
    of unregularised_filter_denominator (unreg_filter_denom is the so here named denominator/delta_beta
    of equation 8).

    References
    -----------
    [1] B. Yu, L. Weber, A. Pacureanu, M. Langer, C. Olivier, P. Cloetens, and F. Peyrin,
    "Evaluation of phase retrieval approaches in magnified X-ray phase nano computerized tomography applied to bone tissue",
    Optics Express, Vol 26, No 9, 11110-11124 (2018)
    """

    def __init__(
        self,
        shape,
        geo_pars,
        delta_beta,
        padded_shape="auto",
        padding_mode="reflect",
        translation_vh=None,
        normalize_by_mean=False,
        lim1=1.0e-5,
        lim2=0.2,
        use_rfft=False,
        fftw_num_threads=None,
        fft_num_threads=None,
        logger=None,
    ):
        """
        Initialize a Contrast Transfer Function phase retrieval.

        Parameters
        ----------
        geo_pars: GeoPars
            the geometry description
        delta_beta : float
            the delta/beta ratio
        padded_shape: str or tuple, optional
            Padded image shape, in the form (num_rows, num_columns) i.e (vertical, horizontal).
            By default, it is twice the image shape.
        padding_mode: str
            Padding mode. It must be valid for the numpy.pad function
        translation_vh: array, optional
            Shift in the form (y, x). It is used to perform a translation of the image before applying the CTF filter.
        normalize_by_mean: bool
            Whether to divide the (padded) image with its mean before applying the CTF filter.
        lim1: float >0
            the regulariser strenght at low  frequencies
        lim2: float >0
            the regulariser strenght at high frequencies
        use_rfft: bool, optional
            Whether to use real-to-complex (R2C) FFT instead of usual complex-to-complex (C2C).
        fftw_num_threads: bool or None or int, optional
            DEPRECATED - please use fft_num_threads instead.
        fft_num_threads: bool or None or int, optional
            Number of threads to use for FFT.
            If a number is provided: number of threads to use for FFT.
            You can pass a negative number to use N - fft_num_threads cores.
        logger: optional
            a logger object
        """
        self.logger = LoggerOrPrint(logger)
        if not isinstance(geo_pars, GeoPars):
            raise ValueError("Expected GeoPars instance for 'geo_pars' parameter")
        self.geo_pars = geo_pars
        self._calc_shape(shape, padded_shape, padding_mode)
        self.delta_beta = delta_beta

        # COMPAT.
        if fftw_num_threads is not None:
            deprecation_warning("'fftw_num_threads' is replaced with 'fft_num_threads'", func_name="ctf_fftw")
            fft_num_threads = fftw_num_threads
        # ---

        self.lim = None
        self.lim1 = lim1
        self.lim2 = lim2
        self.normalize_by_mean = normalize_by_mean
        self.translation_vh = translation_vh
        self._setup_fft(use_rfft, fft_num_threads)
        self._get_ctf_filter()

    def _calc_shape(self, shape, padded_shape, padding_mode):
        if np.isscalar(shape):
            shape = (shape, shape)
        else:
            assert len(shape) == 2
        self.shape = shape
        if padded_shape is None or padded_shape is False:
            padded_shape = self.shape  # no padding
        elif isinstance(padded_shape, (tuple, list, np.ndarray)):
            pass
        elif padded_shape == "auto":
            padded_shape = (2 * self.shape[0], 2 * self.shape[1])
        self.shape_padded = tuple(padded_shape)
        self.padding_mode = padding_mode

    def _setup_fft(self, use_rfft, fft_num_threads):
        self.use_rfft = use_rfft
        self._fft_func = rfft2 if use_rfft else fft2
        self._ifft_func = irfft2 if use_rfft else ifft2
        self.fft_num_threads = get_num_threads(fft_num_threads)

    def _get_ctf_filter(self):
        """
        The parameter "length_scale" was mentioned, in the octave code,
        as a rescaling length scale, which is meant to avoid
        having too big or too small numbers.
        From the mathematical point of view, it is infact completely trasparent:
        its action is on fsamplex, fsampley and  betash, betasv.
        But these latters ( beta's)  are multiplied by  the formers (fsample's) so that
        "length_scale" mathematically disappears, however in case of simple precision float
        the exponent of a float number ranges from -38 to 38, and one could approach it
        as an example by taking the square of a very small number ( 1.0e-19), and losing significant bits
        in he mantissa or getting zero, or the square of a big number, thus generting inf
        Althought the values involved in our x-ray regimes seems safe, with respect to these problems,
        this length_scale parameters does not hurt.
        """
        padded_img_shape = self.shape_padded
        fsample_vh = np.array(
            [
                self.geo_pars.length_scale / self.geo_pars.pix_size_rec_xy[1],
                self.geo_pars.length_scale / self.geo_pars.pix_size_rec_xy[0],
            ]
        )

        if not self.use_rfft:
            ff_index_vh = list(map(np.fft.fftfreq, padded_img_shape))
        else:
            ff_index_vh = [np.fft.fftfreq(padded_img_shape[0]), np.fft.rfftfreq(padded_img_shape[1])]

        # if padded_img_shape[1]%2 == 0 : # change to holotomo_slave indexing (by a transparent 2pi shift)
        #     ff_index_x[ ff_index_x  == -0.5    ] =  +0.5
        # if padded_img_shape[0]%2 == 0 : # change to holotomo_slave indexing (by a transparent 2pi shift)
        #     ff_index_y[ ff_index_y  == -0.5    ] =  +0.5

        frequencies_vh = np.array(
            np.meshgrid(ff_index_vh[0] * fsample_vh[0], ff_index_vh[1] * fsample_vh[1], indexing="ij")
        )

        frequencies_squared_vh = frequencies_vh * frequencies_vh

        """
        ---------------  fresnelnumbers and forward propagators -------------------
        In the limit of parallel beam, z1_h and z1_v would be infinite
        so that the here below distances becomes z2
        which is sample-detector distance.
        """

        if self.geo_pars.z1_vh is not None:
            distances_vh = (self.geo_pars.z1_vh * self.geo_pars.z2) / (self.geo_pars.z1_vh + self.geo_pars.z2)
        else:
            distances_vh = np.array([self.geo_pars.z2, self.geo_pars.z2])

        """
        Citing David Paganin (2002) :
        The intensity I_{R_1}(r⊥,z) at a distance z  of a weakly refracting object illuminated by a
        point source at distance R1  behind the said object, is related to the intensity
        I∞(r⊥,z), which would result from normally
        incident collimated illumination of the same object, by (Pogany et al., 1997):
          I_{R_1}(r⊥,z) = 1/{M^2} I∞(r⊥/M , z/M)
        where M is the magnification.
        This explains the effective distance formula expressed by
        distancesh, distancesv  above.
        ------------------------------------------------------------------------------
        """

        lambda_dist_vh = self.geo_pars.wavelength * distances_vh / (self.geo_pars.length_scale**2)

        """
        ---------------------------------------------------------------------------------------
        -> cut_v at first maximum of ctf largest distance

        In the paraxial expansion of the Fresnel propagator, the phase is equal to
                          1/2 K_{parallel}^2 wavelength * Distance /2 / pi

        When this is equal to a multiple of 2 pi, the effect of propagation disappears
        and we have a singularity in equation 8.

        The sampling in the reciprocal space is done with  a step length of 2*pi*fsamplex,y
          (note: fsamples are the plain inverse of pixel size)
        The first singularity occurs  at a frequence number
                        K_{parallel}/2/pi/ fsamplex

         for K_{parallel}^2 = 2 * (2pi)^2 /( wavelength * distance  )

         which would correspond to
                   K_{parallel}/2/pi/ fsamplex  =  sqrt( 2/wavelength*distance) / fsample

        Question:( why the factor 2 appear at the denominator in the square root below?)
        Answer : the below defined cut corresponds to the first maximum of the denominator,
        before arriving to the first zero. In this way the regularisation
        is already at ( almost) full strenght on the first pole.ation
        ## is already at ( almost) full strenght on the first pole.
        """
        self.cut_v = math.sqrt(1.0 / 2 / lambda_dist_vh[0]) / fsample_vh[0]
        self.cut_v = min(self.cut_v, 0.5)

        self.logger.debug("Normalized cut-off = {:5.3f}".format(self.cut_v))

        self.r = fourier_filters.get_lowpass_filter(
            padded_img_shape,
            cutoff_par=(
                0.5 / (self.cut_v + 1.0 / padded_img_shape[0]),
                0.01 * padded_img_shape[0] / (1 + self.cut_v * padded_img_shape[0]),
            ),
            use_rfft=self.use_rfft,
        )
        self.r /= self.r[0, 0]

        self.lim = self.lim1 * self.r + self.lim2 * (1 - self.r)

        # more methods exist in the original code, and they are initialized starting from here
        # (ht_app1, ht_app2... ht_app7)

        fresnel_phase = (
            np.pi * lambda_dist_vh[1] * frequencies_squared_vh[1]
            + np.pi * lambda_dist_vh[0] * frequencies_squared_vh[0]
        )

        if self.delta_beta:
            unreg_filter_denom = np.sin(fresnel_phase) + (1.0 / self.delta_beta) * np.cos(fresnel_phase)
        else:
            unreg_filter_denom = np.sin(fresnel_phase)
        self.unreg_filter_denom = unreg_filter_denom.astype(np.float32)
        self._ctf_filter_denom = (2 * self.unreg_filter_denom * self.unreg_filter_denom + self.lim).astype(np.complex64)

    def _apply_filter(self, img):
        img_f = self._fft_func(img, workers=self.fft_num_threads)
        img_f *= self.unreg_filter_denom

        unreg_filter_denom_0_mean = self.unreg_filter_denom[0, 0]
        nf, mf = img.shape

        # here it is assumed that the average of img is 1 and the DC component is removed
        img_f[0, 0] -= nf * mf * unreg_filter_denom_0_mean

        ## formula 8, with regularisation to stay at a safe distance from the poles
        img_f /= self._ctf_filter_denom
        ph = self._ifft_func(img_f, workers=self.fft_num_threads).real
        return ph

    def retrieve_phase(self, img, output=None):
        """
        Apply the CTF filter to retrieve the phase.

        Parameters
        ----------
        img: np.ndarray
            Projection image. It must have been already flat-fielded.

        Returns
        --------
        ph: numpy.ndarray
            Phase image
        """
        padded_img = pad_interpolate(
            img, self.shape_padded, translation_vh=self.translation_vh, padding_mode=self.padding_mode
        )
        if self.normalize_by_mean:
            padded_img /= padded_img.mean()
        phase_img = self._apply_filter(padded_img)
        res = recut(phase_img, img.shape)
        if output is not None:
            output[:, :] = res[:, :]
            return output
        return res

    __call__ = retrieve_phase


CtfFilter = CTFPhaseRetrieval



from .cli_configs import BootstrapStitchingConfig
from ..pipeline.config import generate_nabu_configfile
from ..stitching.config import (
    get_default_stitching_config,
    SECTIONS_COMMENTS as _SECTIONS_COMMENTS,
    INPUT_DATASETS_FIELD as _INPUT_DATASETS_FIELD,
    INPUTS_SECTION as _INPUTS_SECTION,
)
from .utils import parse_params_values
from tomoscan.factory import Factory
from tomoscan.esrf.volume.utils import guess_volumes


def guess_tomo_objects(my_str: str) -> tuple:
    """
    try to find some tomo object from a string.
    The string can be either related to a volume or a scan and can be an identifier or a filfe/folder path

    :param str my_str: string related to the tomo object
    :return: a tuple of tomo objects either instance of VolumeBase or TomoScanBase
    :rtype: tuple
    """
    try:
        # create_tomo_object_from_identifier will raise an exception is the string does not match an identifier
        return (Factory.create_tomo_object_from_identifier(my_str),)
    except Exception:
        pass

    try:
        volumes = guess_volumes(my_str)
    except Exception:
        pass
    else:
        if len(volumes) > 0:
            return volumes

    try:
        return Factory.create_scan_objects(my_str)
    except Exception:
        return tuple()


def bootstrap_stitching():
    args = parse_params_values(
        BootstrapStitchingConfig,
        parser_description="Initialize a 'nabu-stitching' configuration file",
    )

    prefilled_values = {}

    datasets_as_str = args.get("datasets", None)
    datasets = []
    for dataset in datasets_as_str:
        datasets.extend(guess_tomo_objects(dataset))

    if len(datasets) > 0:
        prefilled_values = {
            _INPUTS_SECTION: {_INPUT_DATASETS_FIELD: [dataset.get_identifier().to_str() for dataset in datasets]}
        }

    generate_nabu_configfile(
        fname=args["output"],
        default_config=get_default_stitching_config(args["stitching_type"]),
        comments=True,
        sections_comments=_SECTIONS_COMMENTS,
        options_level=args["level"],
        prefilled_values=prefilled_values,
    )
    return 0



from nabu.stitching.stitcher.pre_processing import PreProcessingStitching
from nabu.stitching.stitcher.post_processing import PostProcessingStitching
from .dumper import PreProcessingStitchingDumper, PostProcessingStitchingDumperNoDD, PostProcessingStitchingDumper
from nabu.stitching.stitcher.single_axis import _SingleAxisMetaClass


class PreProcessingZStitcher(
    PreProcessingStitching,
    dumper_cls=PreProcessingStitchingDumper,
    axis=0,
):

    def check_inputs(self):
        """
        insure input data is coherent
        """
        super().check_inputs()

        for scan_0, scan_1 in zip(self.series[0:-1], self.series[1:]):
            if scan_0.dim_1 != scan_1.dim_1:
                raise ValueError(
                    f"projections width are expected to be the same. Not the case for {scan_0} ({scan_0.dim_1} and {scan_1} ({scan_1.dim_1}))"
                )


class PostProcessingZStitcher(
    PostProcessingStitching,
    metaclass=_SingleAxisMetaClass,
    dumper_cls=PostProcessingStitchingDumper,
    axis=0,
):
    @property
    def serie_label(self) -> str:
        return "z-serie"


class PostProcessingZStitcherNoDD(
    PostProcessingStitching,
    metaclass=_SingleAxisMetaClass,
    dumper_cls=PostProcessingStitchingDumperNoDD,
    axis=0,
):
    @property
    def serie_label(self) -> str:
        return "z-serie"



from typing import Union

from tomoscan.identifier import BaseIdentifier
from nabu.stitching.stitcher.z_stitcher import PreProcessingZStitcher as PreProcessZStitcher
from nabu.stitching.stitcher.z_stitcher import (
    PostProcessingZStitcher as PostProcessZStitcher,
    PostProcessingZStitcherNoDD as PostProcessZStitcherNoDD,
)
from nabu.stitching.config import (
    PreProcessedZStitchingConfiguration,
    PostProcessedZStitchingConfiguration,
)


def z_stitching(
    configuration: Union[PreProcessedZStitchingConfiguration, PostProcessedZStitchingConfiguration], progress=None
) -> BaseIdentifier:
    """
    Apply stitching from provided configuration. Along axis 0 (aka z)
    Return a DataUrl with the created NXtomo or Volume

    like:
                    axis 0
                      ^
                      |
    x-ray             |
    -------->          ------> axis 2
                     /
                    /
                     axis 1
    """
    stitcher = None
    assert configuration.axis is not None
    if isinstance(configuration, PreProcessedZStitchingConfiguration):
        if configuration.axis == 0:
            stitcher = PreProcessZStitcher(configuration=configuration, progress=progress)
    elif isinstance(configuration, PostProcessedZStitchingConfiguration):
        assert configuration.axis == 0
        if configuration.duplicate_data:
            stitcher = PostProcessZStitcher(configuration=configuration, progress=progress)
        else:
            stitcher = PostProcessZStitcherNoDD(configuration=configuration, progress=progress)

    if stitcher is None:
        raise TypeError(
            f"configuration is expected to be in {(PreProcessedZStitchingConfiguration, PostProcessedZStitchingConfiguration)}. {type(configuration)} provided"
        )
    return stitcher.stitch()



import numpy as np
from ..utils import check_supported

try:
    import pyopencl as cl
    import pyopencl.array as parray

    __has_pyopencl__ = True
    __pyopencl_error_msg__ = None
except ImportError as err:
    __has_pyopencl__ = False
    __pyopencl_error_msg__ = str(err)
from ..resources.gpu import GPUDescription


def get_opencl_devices(
    device_type,
    vendor=None,
    name=None,
    order_by="global_mem_size",
    exclude_platforms=None,
    exclude_vendors=None,
    prefer_GPU=True,
):
    """
    List available OpenCL devices.

    Parameters
    ----------
    device_type: str
        Type of device, can be "CPU", "GPU" or "all".
    vendor: str, optional
        Filter devices by vendor, eg. "NVIDIA"
    name: Filter devices by names, eg. "GeForce RTX 3080"
    order_by: str, optional
        Order results in decreasing order of this value. Default is to sort by global memory size.
    exclude_platforms: str, optional
        Platforms to exclude, eg. "Portable Computing Language"
    exclude_vendors: str, optional
        Vendors to be excluded
    prefer_GPU: bool, optional
        Whether to put GPUs on top of the returned list, regardless of the "order_by" parameter.
        This can be useful when sorting by global memory size, as CPUs often have a bigger memory size.

    Returns
    -------
    devices: list of pyopencl.Device
        List of OpenCL devices matching the criteria, and ordered by the 'order_by' parameter.
        The list may be empty.

    """
    exclude_platforms = exclude_platforms or []
    exclude_vendors = exclude_vendors or []
    dev_type = {
        "cpu": cl.device_type.CPU,
        "gpu": cl.device_type.GPU,
        "accelerator": cl.device_type.ACCELERATOR,
        "all": cl.device_type.ALL,
        "any": cl.device_type.ALL,
    }
    device_type = device_type.lower()
    check_supported(device_type, dev_type.keys(), "device_type")

    devices = []
    for platform in cl.get_platforms():
        if vendor is not None and vendor.lower() not in platform.vendor.lower():
            continue
        if any(excluded_platform.lower() in platform.name.lower() for excluded_platform in exclude_platforms):
            continue
        if any(excluded_vendor.lower() in platform.vendor.lower() for excluded_vendor in exclude_vendors):
            continue
        for device in platform.get_devices():
            if device.type & dev_type[device_type] == 0:
                continue
            if name is not None and name.lower() not in device.name.lower():
                continue
            devices.append(device)
    if order_by is not None:
        devices = sorted(devices, key=lambda dev: getattr(dev, order_by), reverse=True)
    if prefer_GPU:
        # put GPUs devices on top of the list
        devices = [dev for dev in devices if dev.type & dev_type["gpu"] > 0] + [
            dev for dev in devices if dev.type & dev_type["gpu"] == 0
        ]
    return devices


def usable_opencl_devices():
    """
    Test the available OpenCL platforms/devices.

    Returns
    --------
    platforms: dict
        Dictionary where the key is the platform name, and the value is a list
        of `silx.opencl.common.Device` object.
    """
    platforms = {}
    for platform in cl.get_platforms():
        platforms[platform.name] = platform.get_devices()
    return platforms


def detect_opencl_gpus():
    """
    Get the available OpenCL-compatible GPUs.

    Returns
    --------
    gpus: dict
        Nested dictionary where the keys are OpenCL platform names,
        values are dictionary of GPU IDs and `silx.opencl.common.Device` object.
    error_msg: str
        In the case where there is an error, the message is returned in this item.
        Otherwise, it is a None object.
    """
    gpus = {}
    error_msg = None
    if not (__has_pyopencl__):
        return {}, __pyopencl_error_msg__
    try:
        platforms = usable_opencl_devices()
    except Exception as exc:
        error_msg = str(exc)
    if error_msg is not None:
        return {}, error_msg
    for platform_name, devices in platforms.items():
        for d_id, device in enumerate(devices):
            if device.type == cl.device_type.GPU:  # and bool(device.available):
                if platform_name not in gpus:
                    gpus[platform_name] = {}
                gpus[platform_name][d_id] = device
    return gpus, None


def collect_opencl_gpus():
    """
    Return a dictionary of platforms and brief description of each OpenCL-compatible
    GPU with a few fields
    """
    gpus, error_msg = detect_opencl_gpus()
    if error_msg is not None:
        return None
    opencl_gpus = {}
    for platform, gpus in gpus.items():
        for gpu_id, gpu in gpus.items():
            if platform not in opencl_gpus:
                opencl_gpus[platform] = {}
            opencl_gpus[platform][gpu_id] = GPUDescription(gpu, device_id=gpu_id).get_dict()
            opencl_gpus[platform][gpu_id]["platform"] = platform
    return opencl_gpus


def collect_opencl_cpus():
    """
    Return a dictionary of platforms and brief description of each OpenCL-compatible
    CPU with a few fields
    """
    opencl_cpus = {}
    platforms = usable_opencl_devices()
    for platform, devices in platforms.items():
        if "cuda" in platform.lower():
            continue
        opencl_cpus[platform] = {}
        for device_id, device in enumerate(devices):  # device_id might be inaccurate
            if device.type != cl.device_type.CPU:
                continue
            opencl_cpus[platform][device_id] = GPUDescription(device).get_dict()
            opencl_cpus[platform][device_id]["platform"] = platform
    return opencl_cpus


def get_opencl_context(device_type, **kwargs):
    """
    Create an OpenCL context. Please refer to 'get_opencl_devices' documentation
    """
    devices = get_opencl_devices(device_type, **kwargs)
    if devices == []:
        raise RuntimeError("No OpenCL device found for device_type='%s' and %s" % (device_type, str(kwargs)))
    return cl.Context([devices[0]])


def replace_array_memory(arr, new_shape):
    """
    Replace the underlying buffer data of a  `pyopencl.array.Array`.
    This function is dangerous !
    It should merely be used to clear memory, the array should not be used afterwise.
    """
    arr.data.release()
    arr.base_data = cl.Buffer(arr.context, cl.mem_flags.READ_WRITE, np.prod(new_shape) * arr.dtype.itemsize)
    arr.shape = new_shape
    # strides seems to be updated by pyopencl
    return arr


def pick_opencl_cpu_platform(opencl_cpus):
    """
    Pick the best OpenCL implementation for the opencl cpu.
    This function assume that there is only one opencl-enabled CPU on the
    current machine, but there might be several OpenCL implementations/vendors.


    Parameters
    ----------
    opencl_cpus: dict
        Dictionary with the available opencl-enabled CPUs.
        Usually obtained with collect_opencl_cpus().

    Returns
    -------
    cpu: dict
        A dictionary describing the CPU.
    """
    if len(opencl_cpus) == 0:
        raise ValueError("No CPU to pick")
    name2device = {}
    for platform, devices in opencl_cpus.items():
        for device_id, device_desc in devices.items():
            name2device.setdefault(device_desc["name"], [])
            name2device[device_desc["name"]].append(platform)
    if len(name2device) > 1:
        raise ValueError("Expected at most one CPU but got %d: %s" % (len(name2device), list(name2device.keys())))
    cpu_name = list(name2device.keys())[0]
    platforms = name2device[cpu_name]
    # Several platforms for the same CPU
    res = opencl_cpus[platforms[0]]
    if len(platforms) > 1:
        if "intel" in cpu_name.lower():
            for platform in platforms:
                if "intel" in platform.lower():
                    res = opencl_cpus[platform]
    #
    return res[list(res.keys())[0]]


def allocate_texture(ctx, shape, support_1D=False):
    """
    Allocate an OpenCL image ("texture").

    Parameters
    ----------
    ctx: OpenCL context
        OpenCL context
    shape: tuple of int
        Shape of the image. Note that pyopencl and OpenCL < 1.2
        do not support 1D images, so 1D images are handled as 2D with one row
    support_1D: bool, optional
         force the image to be 1D if the shape has only one dim
    """
    if len(shape) == 1 and not (support_1D):
        shape = (1,) + shape
    return cl.Image(
        ctx,
        cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
        cl.ImageFormat(cl.channel_order.INTENSITY, cl.channel_type.FLOAT),
        hostbuf=np.zeros(shape[::-1], dtype=np.float32),
    )


def check_textures_availability(ctx):
    """
    Check whether textures are supported on the current OpenCL context.
    """
    try:
        dummy_texture = allocate_texture(ctx, (16, 16))
        # Need to further access some attributes (pocl)
        dummy_height = dummy_texture.height
        textures_available = True
        del dummy_texture, dummy_height
    except (cl.RuntimeError, cl.LogicError):
        textures_available = False
    # Nvidia Fermi GPUs (compute capability 2.X) do not support opencl read_imagef
    # There is no way to detect this until a kernel is compiled
    try:
        cc = ctx.devices[0].compute_capability_major_nv
        textures_available &= cc >= 3
    except (cl.LogicError, AttributeError):  # probably not a Nvidia GPU
        pass
    #
    return textures_available


def copy_to_texture(queue, dst_texture, src_array, dtype=np.float32):
    shape = src_array.shape
    if isinstance(src_array, parray.Array):
        return cl.enqueue_copy(queue, dst_texture, src_array.data, offset=0, origin=(0, 0), region=shape[::-1])
    elif isinstance(src_array, np.ndarray):
        if not (src_array.flags["C_CONTIGUOUS"] and src_array.dtype == dtype):
            src_array = np.ascontiguousarray(src_array, dtype=dtype)
        return cl.enqueue_copy(queue, dst_texture, src_array, origin=(0, 0), region=shape[::-1])
    else:
        raise ValueError("Unknown source array type")



try:
    from skimage.transform import rotate

    __have__skimage__ = True
except ImportError:
    __have__skimage__ = False


class Rotation:
    supported_modes = {
        "constant": "constant",
        "zeros": "constant",
        "edge": "edge",
        "edges": "edge",
        "symmetric": "symmetric",
        "sym": "symmetric",
        "reflect": "reflect",
        "wrap": "wrap",
        "periodic": "wrap",
    }

    def __init__(self, shape, angle, center=None, mode="edge", reshape=False, **sk_kwargs):
        """
        Initiate a Rotation object.

        Parameters
        ----------
        shape: tuple of int
            Shape of the images to process
        angle: float
            Rotation angle in DEGREES
        center: tuple of float, optional
            Coordinates of the center of rotation, in the format (X, Y) (mind the non-python
            convention !).
            Default is ((Nx - 1)/2.0, (Ny - 1)/2.0)
        mode: str, optional
            Padding mode. Default is "edge".
        reshape: bool, optional


        Other Parameters
        -----------------
        All the other parameters are passed directly to scikit image 'rotate' function:
        order, cval, clip, preserve_range.
        """
        self.shape = shape
        self.angle = angle
        self.center = center
        self.mode = mode
        self.reshape = reshape
        self.sk_kwargs = sk_kwargs

    def rotate(self, img, output=None):
        if not __have__skimage__:
            raise ValueError("scikit-image is needed for using rotate()")
        res = rotate(img, self.angle, resize=self.reshape, center=self.center, mode=self.mode, **self.sk_kwargs)
        if output is not None:
            output[:] = res[:]
            return output
        else:
            return res

    __call__ = rotate



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import os
import sys
import logging
from argparse import RawTextHelpFormatter
import numpy
from silx.io.url import DataUrl
from tomoscan.esrf.volume.utils import guess_volumes
from tomoscan.factory import Factory
from tomoscan.esrf.volume import (
    EDFVolume,
    HDF5Volume,
    JP2KVolume,
    MultiTIFFVolume,
    TIFFVolume,
)
from nabu.io.cast_volume import (
    RESCALE_MAX_PERCENTILE,
    RESCALE_MIN_PERCENTILE,
    cast_volume,
    get_default_output_volume,
)
from nabu.pipeline.params import files_formats
from nabu.utils import convert_str_to_tuple
from nabu.io.cast_volume import _min_max_from_histo


_logger = logging.getLogger(__name__)


def main(argv=None):
    if argv is None:
        argv = sys.argv
    _volume_url_helps = "\n".join(
        [
            f"- {(volume.__name__).ljust(15)}: {volume.example_defined_from_str_identifier()}"
            for volume in (
                EDFVolume,
                HDF5Volume,
                JP2KVolume,
                MultiTIFFVolume,
                TIFFVolume,
            )
        ]
    )

    volume_help = f"""To define a volume you can either provide: \n
    * an url (recommanded way) - see details lower \n
    * a path. For hdf5 and multitiff we expect a file path. For edf, tif and jp2k we expect a folder path. In this case we will try to deduce the Volume from it. \n 
    url must be defined like: \n{_volume_url_helps}
    """

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
    parser.add_argument(
        "input_volume",
        help=f"input volume. {volume_help}",
    )
    parser.add_argument(
        "--output-data-type",
        help="output data type. Valid value are numpy default types name like (uint8, uint16, int8, int16, int32, float32, float64)",
        default="uint16",
    )
    parser.add_argument(
        "--output_volume",
        help=f"output volume. Must be provided if 'output_type' isn't. Must looks like: \n{volume_help}",
        default=None,
    )
    parser.add_argument(
        "--output_type",
        help=f"output type. Must be provided if 'output_volume' isn't. Valid values are {tuple(files_formats.keys())}",
        default=None,
    )
    parser.add_argument(
        "--data_min",
        help=f"value to clamp to volume cast new min. Any lower value will also be clamp to this value.",
        default=None,
    )
    parser.add_argument(
        "--data_max",
        help=f"value to clamp to volume cast new max. Any higher value will also be clamp to this value.",
        default=None,
    )
    parser.add_argument(
        "--rescale_min_percentile",
        help=f"used to determine data_min if not provided. Expected as percentage. Default is {RESCALE_MIN_PERCENTILE}%%",
        default=RESCALE_MIN_PERCENTILE,
    )
    parser.add_argument(
        "--rescale_max_percentile",
        help=f"used to determine data_max if not provided. Expected as percentage. Default is {RESCALE_MAX_PERCENTILE}%%",
        default=RESCALE_MAX_PERCENTILE,
    )
    parser.add_argument(
        "--overwrite",
        dest="overwrite",
        action="store_true",
        default=False,
        help="Overwrite file or dataset if exists",
    )
    parser.add_argument(
        "--compression-ratios",
        dest="compression_ratios",
        default=None,
        help="Define compression ratios for jp2k. Expected as a list like [20, 10, 1] for [quality layer 1, quality layer 2, quality layer 3]... Pass parameter to glymur. See https://glymur.readthedocs.io/en/latest/how_do_i.html#write-images-with-different-compression-ratios-for-different-layers for more details",
    )
    parser.add_argument(
        "--histogram-url",
        dest="histogram_url",
        default=None,
        help="Provide url to the histogram - like: '/{path}/my_file.hdf5?path/to/my/data' with my_file.hdf5 is the file containing the histogram. Located under 'path'. And 'path/to/my/data' is the location of the HDF5 dataset",
    )

    options = parser.parse_args(argv[1:])

    # handle input volume
    if os.path.exists(options.input_volume):
        volumes = guess_volumes(options.input_volume)

        def is_not_histogram(vol_identifier):
            return not (hasattr(vol_identifier, "data_path") and vol_identifier.data_path.endswith("histogram"))

        volumes = tuple(filter(is_not_histogram, volumes))
        if len(volumes) == 0:
            _logger.error(f"no valid volume found in {options.input_volume}")
            exit(1)
        elif len(volumes) > 1:
            _logger.error(
                f"found several volume from {options.input_volume}. Please provide one full url from {[volume.get_identifier() for volume in volumes]}"
            )
        else:
            input_volume = volumes[0]
    else:
        try:
            input_volume = Factory.create_tomo_object_from_identifier(options.input_volume)
        except Exception as e:
            raise ValueError(f"Fail to build input volume from url {options.input_volume}") from e

    # handle output format
    output_format = files_formats.get(options.output_type, None)

    # handle output volume
    if options.output_volume is not None:
        # if an url is provided
        if ":" in options.output_volume:
            try:
                output_volume = Factory.create_tomo_object_from_identifier(options.output_volume)
            except Exception as e:
                raise ValueError(f"Fail to build output volume from {options.output_volume}") from e

            if output_format is not None:
                if not (
                    isinstance(output_volume, EDFVolume)
                    and output_format == "edf"
                    or isinstance(output_format, HDF5Volume)
                    and output_format == "hdf5"
                    or isinstance(output_format, JP2KVolume)
                    and output_format == "jp2"
                    or isinstance(output_format, (TIFFVolume, MultiTIFFVolume))
                    and output_format == "tiff"
                ):
                    raise ValueError(
                        "Requested 'output_type' and output volume url are incoherent. 'output_type' is optional when url provided"
                    )
        else:
            path_extension = os.path.splitext(options.output_volume)[-1]
            if path_extension == "":
                # if a folder ha sbeen provided we try to create a volume from this path and the output format
                if output_format == "tiff":
                    output_volume = TIFFVolume(
                        folder=options.output_volume,
                    )
                elif output_format == "edf":
                    output_volume = EDFVolume(
                        folder=options.output_volume,
                    )
                elif output_format == "jp2":
                    output_volume = JP2KVolume(
                        folder=options.output_volume,
                    )
                else:
                    raise ValueError(
                        f"Unable to deduce an output volume from {options.output_volume} and output format {output_format}. Please provide an output_volume as an url"
                    )
            else:
                # if a file path_has been provided
                if path_extension.lower() in ("tif", "tiff") and output_format in (
                    None,
                    "tiff",
                ):
                    output_volume = MultiTIFFVolume(
                        file_path=options.output_volume,
                    )
                elif path_extension.lower() in (
                    "h5",
                    "nx",
                    "nexus",
                    "hdf",
                    "hdf5",
                ) and output_format in (None, "hdf5"):
                    output_volume = HDF5Volume(
                        file_path=options.output_volume,
                        data_path="volume",
                    )
                else:
                    raise ValueError(
                        f"Unable to deduce an output volume from {options.output_volume} and output format {output_format}. Please provide an output_volume as an url"
                    )

    elif options.output_type is None:
        raise ValueError("'output_type' or 'output_volume' is expected")
    else:
        output_volume = get_default_output_volume(
            input_volume=input_volume, output_type=output_format  # pylint: disable=E0606
        )
    try:
        output_data_type = numpy.dtype(getattr(numpy, options.output_data_type))
    except Exception as e:
        raise ValueError(f"Unable to get output data type from {options.output_data_type}") from e

    # get data_min and data_max
    data_min = options.data_min
    if data_min is not None:
        data_min = float(data_min)
    data_max = options.data_max
    if data_max is not None:
        data_max = float(data_max)

    # get rescale_min_percentile and rescale_min_percentile
    rescale_min_percentile = options.rescale_min_percentile

    def clean_percentiles_str(percentile):
        # remove ' char
        percentile = percentile.rstrip("'").lstrip("'")
        # remove " char
        percentile = percentile.rstrip('"').lstrip('"')
        # remove % char
        return percentile.rstrip("%")

    if isinstance(rescale_min_percentile, str):
        rescale_min_percentile = float(clean_percentiles_str(rescale_min_percentile))
    rescale_max_percentile = options.rescale_max_percentile
    if isinstance(rescale_min_percentile, str):
        rescale_max_percentile = float(clean_percentiles_str(rescale_max_percentile))
    assert rescale_min_percentile is not None, "rescale_min_percentile should be an int"
    assert rescale_max_percentile is not None, "rescale_max_percentile should be an int"

    # handle histogram and data_min, data_max
    if options.histogram_url is not None:
        if data_min is not None or data_max is not None:
            raise ValueError("Both histogram url and data min/max are provided. Don't know which one to take")
        else:
            if not options.histogram_url.startswith("silx:"):
                options.histogram_url = "silx:" + options.histogram_url
            histogram_url = DataUrl(path=options.histogram_url)
            data_min, data_max = _min_max_from_histo(
                url=histogram_url,
                rescale_min_percentile=rescale_min_percentile,
                rescale_max_percentile=rescale_max_percentile,
            )

    # update output volume from options
    output_volume.overwrite = options.overwrite
    if options.compression_ratios is not None:
        output_volume.cratios = list([int(value) for value in convert_str_to_tuple(options.compression_ratios)])

    # do volume casting
    cast_volume(
        input_volume=input_volume,
        output_volume=output_volume,
        output_data_type=output_data_type,
        data_min=data_min,
        data_max=data_max,
        rescale_min_percentile=rescale_min_percentile,
        rescale_max_percentile=rescale_max_percentile,
    )
    exit(0)


if __name__ == "__main__":
    main(sys.argv)
